# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""\
XDR based trajectory files --- :mod:`MDAnalysis.coordinates.XDR`
================================================================

This module contains helper function and classes to read the XTC and TRR file
formats.

See Also
--------
MDAnalysis.coordinates.XTC: Read and write GROMACS XTC trajectory files.
MDAnalysis.coordinates.TRR: Read and write GROMACS TRR trajectory files.
MDAnalysis.lib.formats.libmdaxdr: Low level xdr format reader
"""

import errno
import numpy as np
from os.path import getctime, getsize, isfile, split, join
import warnings
from filelock import FileLock

from . import base
from ..lib.mdamath import triclinic_box
from ..lib.util import store_init_arguments


def offsets_filename(filename, ending="npz"):
    """Return offset or its lock filename for XDR files.
    For this the filename is appended
    with `_offsets.{ending}`.

    Parameters
    ----------
    filename : str
        filename of trajectory
    ending : str (optional)
        fileending of offsets file

    Returns
    -------
    offset_filename : str

    """
    head, tail = split(filename)
    return join(head, f".{tail}_offsets.{ending}")


def read_numpy_offsets(filename):
    """read offsets into dictionary.

    This assume offsets have been saved using numpy

    Parameters
    ----------
    filename : str
        filename of offsets

    Returns
    -------
    offsets : dict
        dictionary of offsets information

    """
    try:
        return {k: v for k, v in np.load(filename).items()}

    #  `ValueError` is encountered when the offset file is corrupted.
    except (ValueError, IOError):
        warnings.warn(f"Failed to load offsets file {filename}\n")
        return False


class XDRBaseReader(base.ReaderBase):
    """Base class for libmdaxdr file formats xtc and trr

    This class handles integration of XDR based formats into MDAnalysis. The
    XTC and TRR classes only implement `_write_next_frame` and
    `_frame_to_ts`.

    .. _offsets-label:

    Notes
    -----
    XDR based readers store persistent offsets on disk. The offsets are used to
    enable access to random frames efficiently. These offsets will be generated
    automatically the  first time the  trajectory is opened.  Generally offsets
    are stored  in hidden  `*_offsets.npz` files.  Afterwards opening  the same
    file again is fast. It sometimes can happen that the stored offsets get out
    off sync with the trajectory they refer to. For this the offsets also store
    the number of atoms, size of the file and last modification time. If any of
    them change  the offsets are recalculated.  Writing of the offset  file can
    fail when the  directory where the trajectory file resides  is not writable
    or if the  disk is full. In this  case a warning message will  be shown but
    the offsets will nevertheless be used during the lifetime of the trajectory
    Reader. However, the  next time the trajectory is opened,  the offsets will
    have to be rebuilt again.

    .. versionchanged:: 1.0.0
       XDR offsets read from trajectory if offsets file read-in fails
    .. versionchanged:: 2.0.0
       Add a InterProcessLock when generating offsets
    .. versionchanged:: 2.4.0
       Use a direct read into ts attributes
    .. versionchanged:: 2.9.0
       Changed fasteners.InterProcessLock() to filelock.FileLock
    """

    @store_init_arguments
    def __init__(
        self,
        filename,
        convert_units=True,
        sub=None,
        refresh_offsets=False,
        dt=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        filename : str
            trajectory filename
        convert_units : bool (optional)
            convert units to MDAnalysis units
        sub : array_like (optional)
            `sub` is an array of indices to pick out the corresponding
            coordinates and load only them; this requires that the topology
            itself is that of the sub system.
        refresh_offsets : bool (optional)
            force refresh of offsets
        dt : float (optional)
            timestep in MDAnalysis units to load trajectory with;
            if `dt` is ``None``, the time is taken from the xdr file;
            else, the time is set to `dt` * frame
        **kwargs : dict
            General reader arguments.

        """
        super(XDRBaseReader, self).__init__(
            filename, convert_units=convert_units, **kwargs
        )
        self._xdr = self._file(self.filename)

        self._sub = sub
        if self._sub is not None:
            self.n_atoms = len(self._sub)
        else:
            self.n_atoms = self._xdr.n_atoms

        if not refresh_offsets:
            self._load_offsets()
        else:
            self._read_offsets(store=True)
        frame = self._xdr.read()
        try:
            xdr_frame = self._xdr.read()
            if dt is None:
                dt = xdr_frame.time - frame.time
            else:
                self._ts_kwargs["dt"] = dt
            self._xdr.seek(1)
        except StopIteration:
            dt = 0

        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)
        self._frame = 0
        self._frame_to_ts(frame, self.ts)
        # these should only be initialized once
        self.ts.dt = dt
        self.ts.dimensions = triclinic_box(*frame.box)
        if self.convert_units:
            if self.ts.dimensions is not None:
                self.convert_pos_from_native(self.ts.dimensions[:3])

    @classmethod
    def parse_n_atoms(cls, filename, **kwargs):
        with cls._file(filename) as f:
            n_atoms = f.n_atoms
        return n_atoms

    def close(self):
        """close reader"""
        self._xdr.close()

    def _load_offsets(self):
        """load frame offsets from file, reread them from the trajectory if that
        fails. To prevent the competition of generating the same offset file
        from multiple processes, an `InterProcessLock` is used."""
        fname = offsets_filename(self.filename)
        lock_name = offsets_filename(self.filename, ending="lock")

        #  check if the location of the lock is writable.
        try:
            with FileLock(lock_name) as filelock:
                pass
        except OSError as e:
            if isinstance(e, PermissionError) or e.errno == errno.EROFS:
                warnings.warn(
                    f"Cannot write lock/offset file in same location as "
                    f"{self.filename}. Using slow offset calculation."
                )
                self._read_offsets(store=False)
                return
            else:
                raise

        with FileLock(lock_name) as filelock:
            if not isfile(fname):
                self._read_offsets(store=True)
                return

            # if offsets file read correctly, data will be a dictionary of offsets
            # if not, data will be False
            # if False, offsets should be read from the trajectory
            # this warning can be avoided by loading Universe like:
            # u = mda.Universe(data.TPR, data.XTC, refresh_offsets=True)
            # refer to Issue #1893
            data = read_numpy_offsets(fname)
            if not data:
                warnings.warn(
                    f"Reading offsets from {fname} failed, "
                    "reading offsets from trajectory instead.\n"
                    "Consider setting 'refresh_offsets=True' "
                    "when loading your Universe."
                )
                self._read_offsets(store=True)
                return

            ctime_ok = size_ok = n_atoms_ok = False

            try:
                ctime_ok = getctime(self.filename) == data["ctime"]
                size_ok = getsize(self.filename) == data["size"]
                n_atoms_ok = self._xdr.n_atoms == data["n_atoms"]
            except KeyError:
                # we tripped over some old offset formated file
                pass

            if not (ctime_ok and size_ok and n_atoms_ok):
                warnings.warn(
                    "Reload offsets from trajectory\n "
                    "ctime or size or n_atoms did not match"
                )
                self._read_offsets(store=True)
            else:
                self._xdr.set_offsets(data["offsets"])

    def _read_offsets(self, store=False):
        """read frame offsets from trajectory"""
        fname = offsets_filename(self.filename)
        offsets = self._xdr.offsets
        if store:
            ctime = getctime(self.filename)
            size = getsize(self.filename)
            try:
                np.savez(
                    fname,
                    offsets=offsets,
                    size=size,
                    ctime=ctime,
                    n_atoms=self._xdr.n_atoms,
                )
            except Exception as e:
                warnings.warn(f"Couldn't save offsets because: {e}")

    @property
    def n_frames(self):
        """number of frames in trajectory"""
        return len(self._xdr)

    def _reopen(self):
        """reopen trajectory"""
        self.ts.frame = 0
        self._frame = -1
        offsets = self._xdr.offsets.copy()
        self._xdr.close()
        self._xdr.open(self.filename.encode("utf-8"), "r")
        # only restore in case we actually had offsets
        if len(offsets) != 0:
            self._xdr.set_offsets(offsets)

    def _read_frame(self, i):
        """read frame i"""
        self._frame = i - 1
        try:
            self._xdr.seek(i)
            timestep = self._read_next_timestep()
        except IOError:
            warnings.warn("seek failed, recalculating offsets and retrying")
            offsets = self._xdr.calc_offsets()
            self._xdr.set_offsets(offsets)
            self._read_offsets(store=True)
            self._xdr.seek(i)
            timestep = self._read_next_timestep()
        return timestep

    def Writer(self, filename, n_atoms=None, **kwargs):
        """Return writer for trajectory format"""
        if n_atoms is None:
            n_atoms = self.n_atoms
        return self._writer(filename, n_atoms=n_atoms, **kwargs)


class XDRBaseWriter(base.WriterBase):
    """Base class for libmdaxdr file formats xtc and trr"""

    def __init__(
        self, filename, n_atoms, convert_units=True, dt=None, **kwargs
    ):
        """
        Parameters
        ----------
        filename : str
            filename of trajectory
        n_atoms : int
            number of atoms to be written
        convert_units : bool (optional)
            convert from MDAnalysis units to format specific units
        dt : float (optional)
            timestep in MDAnalysis units to write trajectory with;
            if `dt` is ``None``, time for a frame is set from the timestep;
            else, the time for a frame is `dt` * frame
        **kwargs : dict
            General writer arguments
        """
        self.filename = filename
        self._convert_units = convert_units
        self.n_atoms = n_atoms
        self._xdr = self._file(self.filename, "w")
        self._dt = dt

    def close(self):
        """close trajectory"""
        self._xdr.close()

    def __del__(self):
        self.close()



# -*- coding: utf-8 -*-
#
# MDAnalysis documentation build configuration file, created by
# sphinx-quickstart on Mon Sep 27 09:39:55 2010.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import datetime
import MDAnalysis as mda

# Custom MDA Formating
from pybtex.style.formatting.unsrt import Style as UnsrtStyle
from pybtex.style.labels import BaseLabelStyle
from pybtex.plugin import register_plugin

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown
# here.

# make sure sphinx always uses the current branch
sys.path.insert(0, os.path.abspath("../../.."))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx_sitemap",
    "mdanalysis_sphinx_theme",
    "sphinxcontrib.bibtex",
    "sphinx.ext.doctest",
]

bibtex_bibfiles = ["references.bib"]


# Define custom MDA style for references
class KeyLabelStyle(BaseLabelStyle):
    def format_labels(self, entries):
        entry_list = []
        for entry in entries:
            author = str(entry.persons["author"][0]).split(",")[0]
            year = entry.fields["year"]
            entry_list.append(f"{author}{year}")
        return entry_list


class KeyStyle(UnsrtStyle):
    default_label_style = "keylabel"


register_plugin("pybtex.style.labels", "keylabel", KeyLabelStyle)
register_plugin("pybtex.style.formatting", "MDA", KeyStyle)

mathjax_path = "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.0/MathJax.js?config=TeX-AMS-MML_HTMLorMML"

# for sitemap with https://github.com/jdillard/sphinx-sitemap
# This sitemap is correct both for the development and release docs, which
# are both served from docs.mdanalysis.org/$version .
# The docs used to be available automatically at mdanalysis.org/mdanalysis;
# they are now at docs.mdanalysis.org/, which requires a CNAME DNS record
# pointing to mdanalysis.github.io. To change this URL you should change/delete
# the CNAME record for "docs" and update the URL in GitHub settings
site_url = "https://docs.mdanalysis.org/"

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
# 'index' has the advantage that it is immediately picked up by the webserver
master_doc = "index"

# General information about the project.
# (take the list from AUTHORS)
# Ordering: (1) Naveen (2) Elizabeth, then all contributors in alphabetical order
#           (last) Oliver
author_list = mda.__authors__
authors = ", ".join(author_list[:-1]) + ", and " + author_list[-1]
project = "MDAnalysis"
now = datetime.datetime.now()
copyright = "2005-{}, ".format(now.year) + authors

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# Dynamically calculate the version
packageversion = __import__("MDAnalysis").__version__
# The short X.Y version.
# version = '.'.join(packageversion.split('.')[:2])
version = packageversion  # needed for right sitemap.xml URLs
# The full version, including alpha/beta/rc tags.
release = packageversion

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "default"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# to include decorated objects like __init__
autoclass_content = "both"

# to prevent including of member entries in toctree
toc_object_entries = False

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "mdanalysis_sphinx_theme"

extra_nav_links = {}
extra_nav_links["MDAnalysis"] = "https://mdanalysis.org"
extra_nav_links["User guide"] = "https://userguide.mdanalysis.org"
extra_nav_links["MDAKits"] = "https://mdakits.mdanalysis.org/"


html_theme_options = {
    "mda_official": True,
    "extra_nav_links": extra_nav_links,
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_context = {
    "versions_json_url": "https://docs.mdanalysis.org/versions.json"
}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# For RTD theme: custom.css to override theme defaults.
html_static_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# alabaster sidebars
html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",
        "searchbox.html",
    ]
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
html_use_opensearch = "https://docs.mdanalysis.org"

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "MDAnalysisdoc"


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("MDAnalysis.tex", "MDAnalysis Documentation", authors, "manual"),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("mdanalysis", "MDAnalysis Documentation", [authors], 1)]


# -- Options for Epub output ---------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = "MDAnalysis"
epub_author = authors
epub_publisher = "Arizona State University, Tempe, Arizona, USA"
epub_copyright = "2015, " + authors

# The language of the text. It defaults to the language option
# or en if the language is not set.
# epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
# epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
# epub_identifier = ''

# A unique identification for the text.
# epub_uid = ''

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_pre_files = []

# HTML files shat should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_post_files = []

# A list of files that should not be packed into the epub file.
# epub_exclude_files = []

# The depth of the table of contents in toc.ncx.
# epub_tocdepth = 3

# Allow duplicate toc entries.
# epub_tocdup = True


# Configuration for intersphinx: refer to the Python standard library
# and other packages used by MDAnalysis
intersphinx_mapping = {
    "h5py": ("https://docs.h5py.org/en/stable", None),
    "python": ("https://docs.python.org/3/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "gsd": ("https://gsd.readthedocs.io/en/stable/", None),
    "maplotlib": ("https://matplotlib.org/stable/", None),
    "griddataformats": ("https://mdanalysis.org/GridDataFormats/", None),
    "pmda": ("https://mdanalysis.org/pmda/", None),
    "networkx": ("https://networkx.org/documentation/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "parmed": ("https://parmed.github.io/ParmEd/html/", None),
    "rdkit": ("https://rdkit.org/docs/", None),
    "waterdynamics": ("https://www.mdanalysis.org/waterdynamics/", None),
    "pathsimanalysis": ("https://www.mdanalysis.org/PathSimAnalysis/", None),
    "mdahole2": ("https://www.mdanalysis.org/mdahole2/", None),
    "dask": ("https://docs.dask.org/en/stable/", None),
    "imdclient": ("https://imdclient.readthedocs.io/en/stable/", None),
}



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2020 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
HELANAL --- analysis of protein helices
=======================================

:Author: Lily Wang
:Year: 2020
:Copyright: Lesser GNU Public License v2.1+

.. versionadded:: 2.0.0

This module contains code to analyse protein helices using the
HELANAL_ algorithm
([Bansal2000]_ , [Sugeta1967]_ ).

HELANAL_ quantifies the geometry of helices in proteins on the basis of their
Cα atoms. It can determine local structural features such as the local
helical twist and rise, virtual torsion angle, local helix origins and
bending angles between successive local helix axes.

.. _HELANAL: https://pubmed.ncbi.nlm.nih.gov/10798526/

.. [Sugeta1967] Sugeta, H. and Miyazawa, T. 1967. General method for
   calculating helical parameters of polymer chains from bond lengths, bond
   angles and internal rotation angles. *Biopolymers* 5 673 - 679

.. [Bansal2000] Bansal M, Kumar S, Velavan R. 2000.
   HELANAL - A program to characterise helix geometry in proteins.
   *J Biomol Struct Dyn.*  17(5):811-819.


Example use
-----------

You can pass in a single selection::

    import MDAnalysis as mda
    from MDAnalysis.tests.datafiles import PSF, DCD
    from MDAnalysis.analysis import helix_analysis as hel
    u = mda.Universe(PSF, DCD)
    helanal = hel.HELANAL(u, select='name CA and resnum 161-187')
    helanal.run()

All computed properties are available in ``.results``::

    print(helanal.results.summary)

Alternatively, you can analyse several helices at once by passing
in multiple selection strings::

    helanal2 = hel.HELANAL(u, select=('name CA and resnum 100-160',
                                      'name CA and resnum 200-230'))

The :func:`helix_analysis` function will carry out helix analysis on
atom positions, treating each row of coordinates as an alpha-carbon
equivalent::

    hel_xyz = hel.helix_analysis(u.atoms.positions, ref_axis=[0, 0, 1])

    
Classes
-------

.. autoclass:: HELANAL


Functions
---------

.. autofunction:: helix_analysis

.. autofunction:: vector_of_best_fit

.. autofunction:: local_screw_angles


"""

import warnings
import numpy as np

import MDAnalysis as mda
from ..lib import util, mdamath
from .base import AnalysisBase


def vector_of_best_fit(coordinates):
    """Fit vector through the centered coordinates,
    pointing to the first coordinate (i.e. upside-down).

    Parameters
    ----------
    coordinates : :class:`numpy.ndarray` of shape (N, 3)

    Returns
    -------
    :class:`numpy.ndarray` of shape (3,)
        Vector of best fit.
    """
    centered = coordinates - coordinates.mean(axis=0)
    Mt_M = np.matmul(centered.T, centered)
    u, s, vh = np.linalg.svd(Mt_M)
    vector = vh[0]

    # does vector face first local helix origin?
    angle = mdamath.angle(centered[0], vector)
    if angle > np.pi / 2:
        vector *= -1
    return vector


def local_screw_angles(global_axis, ref_axis, helix_directions):
    """
    Cylindrical azimuth angles between the local direction vectors,
    as projected onto the cross-section of the helix, from (-pi, pi].
    The origin (angle=0) is set to the plane of global_axis and ref_axis.

    Parameters
    ----------
    global_axis : :class:`numpy.ndarray` of shape (3,)
        Vector of best fit. Screw angles are calculated perpendicular to
        this axis.
    ref_axis : :class:`numpy.ndarray` of shape (3,)
        Reference length-wise axis. One of the reference vectors is
        orthogonal to this axis.
    helix_directions : :class:`numpy.ndarray` of shape (N, 3)
        array of vectors representing the local direction of each
        helix window.

    Returns
    -------
    :class:`numpy.ndarray` of shape (N,)
        Array of screw angles.
    """
    global_axis = np.asarray(global_axis)
    # normal to the plane of `ref_axis` & `global_axis`
    perp = np.cross(ref_axis, global_axis)
    if not np.any(perp):  # zero when ref_axis, global_axis parallel
        # use random orthogonal vector
        new_ref = [[1, 0, 0], [0, 0, 1]]
        while not np.any(perp) and new_ref:
            perp = np.cross(new_ref.pop(), global_axis)

    # normal for angle to plane of perp and global_axis
    ortho = np.cross(-perp, global_axis)

    # project helix_directions onto global to remove contribution
    norm_global_sq = np.dot(global_axis, global_axis)
    mag_g = np.matmul(global_axis, helix_directions.T) / norm_global_sq
    # projection onto global_axis
    proj_g = mag_g.reshape(-1, 1) @ global_axis.reshape(1, -1)
    # projection onto plane w/o global_axis contribution
    proj_plane = helix_directions - proj_g

    # angles from projection to perp
    refs = np.array([perp, ortho])  # (2, 3)
    norms = _, ortho_norm = np.outer(
        mdamath.pnorm(refs), mdamath.pnorm(proj_plane)
    )
    cos = cos_perp, cos_ortho = np.matmul(refs, proj_plane.T) / norms
    to_perp, to_ortho = np.arccos(np.clip(cos, -1, 1))  # (2, n_vec)
    to_ortho[ortho_norm == 0] = 0  # ?
    to_ortho[cos_perp < 0] *= -1
    to_ortho[to_ortho == -np.pi] = np.pi  # leave 180 alone
    return np.rad2deg(to_ortho)


def helix_analysis(positions, ref_axis=(0, 0, 1)):
    r"""
    Calculate helix properties from atomic coordinates.

    Each property is calculated from a sliding window of 4 atoms,
    from i to i+3. Any property whose name begins with 'local' is a
    property of a sliding window.

    Parameters
    ----------
    positions : :class:`numpy.ndarray` of shape (N, 3)
        Atomic coordinates.
    ref_axis : array-like of length 3, optional
        The reference axis used to calculate the tilt of the vector
        of best fit, and the local screw angles.

    Returns
    -------
    dict with the following keys:
        local_twists : array, shape (N-3,)
            local twist angle from atom i+1 to i+2
        local_nres_per_turn : array, shape (N-3,)
            number of residues per turn, based on local_twist
        local_axes :  array, shape (N-3, 3)
            the length-wise helix axis of the local window
        local_bends : array, shape (N-6,)
            the angles between local helix angles, 3 windows apart
        local_heights : array, shape (N-3,)
            the rise of each local helix
        local_helix_directions : array, shape (N-2, 3)
            the unit vector from each local origin to atom i+1
        local_origins : array, shape (N-2, 3)
            the projected origin for each helix
        all_bends : array, shape (N-3, N-3)
            angles between each local axis
        global_axis : array, shape (3,)
            vector of best fit through origins, pointing at the first origin.
        local_screw_angles : array, shape (N-2,)
            cylindrical azimuth angle to plane of global_axis and ref_axis
    """

    #          ^               ^
    #           \             / bi
    #            \           /
    #         CA_i+2 <----- CA_i+1
    #         /    \       /   ^
    #        /    r \     /     \
    #     V /        \ θ /       \
    #      /          \ /       CA_i
    #     v           origin
    #   CA_i+3
    #
    # V: vectors
    # bi: approximate "bisectors" in plane of screen
    #     Note: not real bisectors, as the vectors aren't normalised
    # θ: local_twists
    # origin: origins
    # local_axes: perpendicular to plane of screen. Orthogonal to "bisectors"

    vectors = positions[1:] - positions[:-1]  # (n_res-1, 3)
    bisectors = vectors[:-1] - vectors[1:]  # (n_res-2, 3)
    bimags = mdamath.pnorm(bisectors)  # (n_res-2,)
    adjacent_mag = bimags[:-1] * bimags[1:]  # (n_res-3,)

    # find angle between bisectors for twist and n_residue/turn
    cos_theta = mdamath.pdot(bisectors[:-1], bisectors[1:]) / adjacent_mag
    cos_theta = np.clip(cos_theta, -1, 1)
    twists = np.arccos(cos_theta)  # (n_res-3,)
    local_twists = np.rad2deg(twists)
    local_nres_per_turn = 2 * np.pi / twists

    # find normal to bisectors for local axes
    cross_bi = np.cross(bisectors[:-1], bisectors[1:])  # (n_res-3, 3)
    local_axes = (cross_bi.T / mdamath.pnorm(cross_bi)).T  # (n_res-3, 3)
    local_axes = np.nan_to_num(local_axes)

    zero_vectors = np.tile(np.any(local_axes, axis=1), (len(local_axes), 1)).T
    # find angles between axes for bends
    bend_theta = np.matmul(local_axes, local_axes.T)  # (n_res-3, n_res-3)
    # set angles to 0 between zero-vectors
    bend_theta = np.where(
        zero_vectors + zero_vectors.T, bend_theta, 1  # (n_res-3, n_res-3)
    )
    bend_matrix = np.rad2deg(np.arccos(np.clip(bend_theta, -1, 1)))
    # local bends are between axes 3 windows apart
    local_bends = np.diagonal(bend_matrix, offset=3)  # (n_res-6,)

    # radius of local cylinder
    radii = (adjacent_mag**0.5) / (2 * (1.0 - cos_theta))  # (n_res-3,)
    # special case: angle b/w bisectors is 0 (should virtually never happen)
    # guesstimate radius = half bisector magnitude
    radii = np.where(cos_theta != 1, radii, (adjacent_mag**0.5) / 2)
    # height of local cylinder
    heights = np.abs(mdamath.pdot(vectors[1:-1], local_axes))  # (n_res-3,)

    local_helix_directions = (bisectors.T / bimags).T  # (n_res-2, 3)

    # get origins by subtracting radius from atom i+1
    origins = positions[1:-1].copy()  # (n_res-2, 3)
    origins[:-1] -= (radii * local_helix_directions[:-1].T).T
    # subtract radius from atom i+2 in last one
    origins[-1] -= radii[-1] * local_helix_directions[-1]

    helix_axes = vector_of_best_fit(origins)
    screw = local_screw_angles(
        helix_axes, np.asarray(ref_axis), local_helix_directions
    )

    results = {
        "local_twists": local_twists,
        "local_nres_per_turn": local_nres_per_turn,
        "local_axes": local_axes,
        "local_bends": local_bends,
        "local_heights": heights,
        "local_helix_directions": local_helix_directions,
        "local_origins": origins,
        "all_bends": bend_matrix,
        "global_axis": helix_axes,
        "local_screw_angles": screw,
    }
    return results


class HELANAL(AnalysisBase):
    r"""
    Perform HELANAL helix analysis on your trajectory.

    Parameters
    ----------
    universe : Universe or AtomGroup
        The Universe or AtomGroup to apply the analysis to.
    select : str or iterable of str, optional
        The selection string to create an atom selection that the HELANAL
        analysis is applied to. Note that HELANAL is designed to work on the
        alpha-carbon atoms of protein residues. If you pass in multiple
        selections, the selections will be analysed separately.
    ref_axis : array-like of length 3, optional
        The reference axis used to calculate the tilt of the vector
        of best fit, and the local screw angles.
    flatten_single_helix : bool, optional
        Whether to flatten results if only one selection is passed.
    split_residue_sequences : bool, optional
        Whether to split the residue sequence into individual helices.
        This keyword only applies if a residue gap is present in the
        AtomGroup generated by a ``select`` string.
        If ``False``, the residues will be analysed as a single helix.
        If ``True``, each group of consecutive residues will be treated
        as a separate helix.
    verbose : bool, optional
        Turn on more logging and debugging.

    Attributes
    ----------
    results.local_twists : array or list of arrays
        The local twist angle from atom i+1 to i+2.
        Each array has shape (n_frames, n_residues-3)
    results.local_nres_per_turn : array or list of arrays
        Number of residues per turn, based on local_twist.
        Each array has shape (n_frames, n_residues-3)
    results.local_axes : array or list of arrays
        The length-wise helix axis of the local window.
        Each array has shape (n_frames, n_residues-3, 3)
    results.local_heights : array or list of arrays
        The rise of each local helix.
        Each array has shape (n_frames, n_residues-3)
    results.local_helix_directions : array or list of arrays
        The unit vector from each local origin to atom i+1.
        Each array has shape (n_frames, n_residues-2, 3)
    results.local_origins :array or list of arrays
        The projected origin for each helix.
        Each array has shape (n_frames, n_residues-2, 3)
    results.local_screw_angles : array or list of arrays
        The local screw angle for each helix.
        Each array has shape (n_frames, n_residues-2)
    results.local_bends : array or list of arrays
        The angles between local helix axes, 3 windows apart.
        Each array has shape (n_frames, n_residues-6)
    results.all_bends : array or list of arrays
        The angles between local helix axes.
        Each array has shape (n_frames, n_residues-3, n_residues-3)
    results.global_axis : array or list of arrays
        The length-wise axis for the overall helix. This points at
        the first helix window in the helix, so it runs opposite to
        the direction of the residue numbers.
        Each array has shape (n_frames, 3)
    results.global_tilts : array or list of arrays
        The angle between the global axis and the reference axis.
        Each array has shape (n_frames,)
    results.summary : dict or list of dicts
        Summary of stats for each property: the mean, the sample
        standard deviation, and the mean absolute deviation.
    """

    # shapes of properties from each frame, relative to n_residues
    attr_shapes = {
        "local_twists": (-3,),
        "local_bends": (-6,),
        "local_heights": (-3,),
        "local_nres_per_turn": (-3,),
        "local_origins": (-2, 3),
        "local_axes": (-3, 3),
        "local_helix_directions": (-2, 3),
        "local_screw_angles": (-2,),
    }

    def __init__(
        self,
        universe,
        select="name CA",
        ref_axis=(0, 0, 1),
        verbose=False,
        flatten_single_helix=True,
        split_residue_sequences=True,
    ):
        super(HELANAL, self).__init__(
            universe.universe.trajectory, verbose=verbose
        )
        selections = util.asiterable(select)
        atomgroups = [universe.select_atoms(s) for s in selections]
        consecutive = []
        # check that residues are consecutive and long enough sequence
        for s, ag in zip(selections, atomgroups):
            groups = util.group_same_or_consecutive_integers(ag.resindices)
            counter = 0
            if len(groups) > 1:
                msg = "Your selection {} has gaps in the residues.".format(s)
                if split_residue_sequences:
                    msg += " Splitting into {} helices.".format(len(groups))
                else:
                    groups = [ag.resindices]
                warnings.warn(msg)

            for g in groups:
                ng = len(g)
                counter += ng
                if ng < 9:
                    warnings.warn(
                        "Fewer than 9 atoms found for helix in "
                        "selection {} with these resindices: {}. "
                        "This sequence will be skipped. HELANAL "
                        "is designed to work on at sequences of "
                        "≥9 residues.".format(s, g)
                    )
                    continue

                ids, counts = np.unique(g, return_counts=True)
                if np.any(counts > 1):
                    dup = ", ".join(map(str, ids[counts > 1]))
                    warnings.warn(
                        "Your selection {} includes multiple atoms "
                        "for residues with these resindices: {}."
                        "HELANAL is designed to work on one alpha-"
                        "carbon per residue.".format(s, dup)
                    )

                consecutive.append(ag[counter - ng : counter])

        self.atomgroups = consecutive
        self.ref_axis = np.asarray(ref_axis)
        self._flatten = flatten_single_helix

    def _zeros_per_frame(self, dims, n_positions=0):
        """Create zero arrays where first 2 dims are n_frames, n_values"""
        first = dims[0] + n_positions
        npdims = (
            self.n_frames,
            first,
        ) + dims[
            1:
        ]  # py27 workaround
        return np.zeros(npdims, dtype=np.float64)

    def _prepare(self):
        n_res = [len(ag) for ag in self.atomgroups]

        for key, dims in self.attr_shapes.items():
            empty = [self._zeros_per_frame(dims, n_positions=n) for n in n_res]
            self.results[key] = empty

        self.results.global_axis = [self._zeros_per_frame((3,)) for n in n_res]
        self.results.all_bends = [
            self._zeros_per_frame((n - 3, n - 3)) for n in n_res
        ]

    def _single_frame(self):
        _f = self._frame_index
        for i, ag in enumerate(self.atomgroups):
            results = helix_analysis(ag.positions, ref_axis=self.ref_axis)
            for key, value in results.items():
                attr = self.results[key]
                attr[i][_f] = value

    def _conclude(self):
        # compute tilt of global axes
        self.results.global_tilts = tilts = []
        norm_ref = (self.ref_axis**2).sum() ** 0.5
        for axes in self.results.global_axis:
            cos = np.matmul(self.ref_axis, axes.T) / (
                mdamath.pnorm(axes) * norm_ref
            )
            cos = np.clip(cos, -1.0, 1.0)
            tilts.append(np.rad2deg(np.arccos(cos)))

        global_attrs = ["global_axis", "global_tilts", "all_bends"]
        attrnames = list(self.attr_shapes.keys()) + global_attrs
        # summarise
        self.results.summary = []
        for i in range(len(self.atomgroups)):
            stats = {}
            for name in attrnames:
                attr = self.results[name]
                mean = attr[i].mean(axis=0)
                dev = np.abs(attr[i] - mean)
                stats[name] = {
                    "mean": mean,
                    "sample_sd": attr[i].std(axis=0, ddof=1),
                    "abs_dev": dev.mean(axis=0),
                }
            self.results.summary.append(stats)

        # flatten?
        if len(self.atomgroups) == 1 and self._flatten:
            for name in attrnames + ["summary"]:
                attr = self.results[name]
                self.results[name] = attr[0]

    def universe_from_origins(self):
        """
        Create MDAnalysis Universe from the local origins.

        Returns
        -------
        Universe or list of Universes
        """
        try:
            origins = self.results.local_origins
        except AttributeError:
            raise ValueError("Call run() before universe_from_origins")

        if not isinstance(origins, list):
            origins = [origins]

        universe = []
        for xyz in origins:
            n_res = xyz.shape[1]
            u = mda.Universe.empty(
                n_res,
                n_residues=n_res,
                atom_resindex=np.arange(n_res),
                trajectory=True,
            ).load_new(xyz)
            universe.append(u)
        if not isinstance(self.results.local_origins, list):
            universe = universe[0]
        return universe



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import warnings

import MDAnalysis as mda
import numpy as np
import pytest
from MDAnalysis.analysis import contacts
from MDAnalysis.analysis.distances import distance_array
from numpy.testing import assert_allclose, assert_array_equal, assert_equal

from MDAnalysisTests.datafiles import (
    DCD,
    PSF,
    TPR,
    XTC,
    contacts_file,
    contacts_villin_folded,
    contacts_villin_unfolded,
)


def test_soft_cut_q():
    # just check some of the extremal points
    assert contacts.soft_cut_q([0], [0]) == 0.5
    assert_allclose(contacts.soft_cut_q([100], [0]), 0, rtol=0, atol=1.5e-7)
    assert_allclose(contacts.soft_cut_q([-100], [0]), 1, rtol=0, atol=1.5e-7)


def test_soft_cut_q_folded():
    u = mda.Universe(contacts_villin_folded)

    contacts_data = np.genfromtxt(contacts_file)
    # indices have been stored 1 indexed
    indices = contacts_data[:, :2].astype(int) - 1

    r = np.linalg.norm(
        u.atoms.positions[indices[:, 0]] - u.atoms.positions[indices[:, 1]],
        axis=1,
    )
    r0 = contacts_data[:, 2]

    beta = 5.0
    lambda_constant = 1.8
    Q = 1 / (1 + np.exp(beta * (r - lambda_constant * r0)))

    assert_allclose(Q.mean(), 1.0, rtol=0, atol=1.5e-3)


def test_soft_cut_q_unfolded():
    u = mda.Universe(contacts_villin_unfolded)

    contacts_data = np.genfromtxt(contacts_file)
    # indices have been stored 1 indexed
    indices = contacts_data[:, :2].astype(int) - 1

    r = np.linalg.norm(
        u.atoms.positions[indices[:, 0]] - u.atoms.positions[indices[:, 1]],
        axis=1,
    )
    r0 = contacts_data[:, 2]

    beta = 5.0
    lambda_constant = 1.8
    Q = 1 / (1 + np.exp(beta * (r - lambda_constant * r0)))

    assert_allclose(Q.mean(), 0.0, rtol=0, atol=1.5e-1)


@pytest.mark.parametrize(
    "r, cutoff, expected_value",
    [
        ([1], 2, 1),
        ([2], 1, 0),
        ([2, 0.5], 1, 0.5),
        ([2, 3], [3, 4], 1),
        ([4, 5], [3, 4], 0),
    ],
)
def test_hard_cut_q(r, cutoff, expected_value):
    # just check some extremal points
    assert contacts.hard_cut_q(r, cutoff) == expected_value


@pytest.mark.parametrize(
    "r, r0, radius, expected_value",
    [([1], None, 2, 1), ([2], None, 1, 0), ([2, 0.5], None, 1, 0.5)],
)
def test_radius_cut_q(r, r0, radius, expected_value):
    # check some extremal points
    assert contacts.radius_cut_q(r, r0, radius) == expected_value


def test_contact_matrix():
    d = np.arange(5)
    radius = np.ones(5) * 2.5

    out = contacts.contact_matrix(d, radius)
    assert_array_equal(out, [True, True, True, False, False])

    # check in-place update
    out = np.empty(out.shape)
    contacts.contact_matrix(d, radius, out=out)
    assert_array_equal(out, [True, True, True, False, False])


def test_new_selection():
    u = mda.Universe(PSF, DCD)
    selections = ("all",)
    sel = contacts._new_selections(u, selections, -1)[0]
    u.trajectory[-1]
    assert_array_equal(sel.positions, u.atoms.positions)


def soft_cut(ref, u, selA, selB, radius=4.5, beta=5.0, lambda_constant=1.8):
    """
    Reference implementation for testing
    """
    # reference groups A and B from selection strings
    refA, refB = ref.select_atoms(selA), ref.select_atoms(selB)

    # 2D float array, reference distances (r0)
    dref = distance_array(refA.positions, refB.positions)

    # 2D bool array, select reference distances that are less than the cutoff
    # radius
    mask = dref < radius

    # group A and B in a trajectory
    grA, grB = u.select_atoms(selA), u.select_atoms(selB)
    results = []

    for ts in u.trajectory:
        d = distance_array(grA.positions, grB.positions)
        r, r0 = d[mask], dref[mask]
        x = 1 / (1 + np.exp(beta * (r - lambda_constant * r0)))

        # average/normalize and append to results
        results.append((ts.time, x.sum() / mask.sum()))

    return np.asarray(results)


class TestContacts(object):
    sel_basic = "(resname ARG LYS) and (name NH* NZ)"
    sel_acidic = "(resname ASP GLU) and (name OE* OD*)"

    @staticmethod
    @pytest.fixture()
    def universe():
        return mda.Universe(PSF, DCD)

    def _run_Contacts(
        self,
        universe,
        client_Contacts,
        start=None,
        stop=None,
        step=None,
        **kwargs,
    ):
        acidic = universe.select_atoms(self.sel_acidic)
        basic = universe.select_atoms(self.sel_basic)
        return contacts.Contacts(
            universe,
            select=(self.sel_acidic, self.sel_basic),
            refgroup=(acidic, basic),
            radius=6.0,
            **kwargs,
        ).run(**client_Contacts, start=start, stop=stop, step=step)

    @pytest.mark.parametrize("seltxt", [sel_acidic, sel_basic])
    def test_select_valid_types(self, universe, seltxt):
        """Test if Contacts._get_atomgroup() can take both string and AtomGroup
        as selections.
        """
        ag = universe.select_atoms(seltxt)

        ag_from_string = contacts.Contacts._get_atomgroup(universe, seltxt)
        ag_from_ag = contacts.Contacts._get_atomgroup(universe, ag)

        assert ag_from_string == ag_from_ag

    def test_contacts_selections(self, universe, client_Contacts):
        """Test if Contacts can take both string and AtomGroup as selections."""
        aga = universe.select_atoms(self.sel_acidic)
        agb = universe.select_atoms(self.sel_basic)

        cag = contacts.Contacts(
            universe, select=(aga, agb), refgroup=(aga, agb)
        )

        csel = contacts.Contacts(
            universe,
            select=(self.sel_acidic, self.sel_basic),
            refgroup=(aga, agb),
        )

        cag.run(**client_Contacts)
        csel.run(**client_Contacts)

        assert cag.grA == csel.grA
        assert cag.grB == csel.grB

    @pytest.mark.parametrize("ag", [1, [2], mda.Universe, "USE UPDATING AG"])
    def test_select_wrong_types(self, universe, ag):
        """Test that Contacts._get_atomgroup(u, sel) fails if sel is of the
        wrong type"""
        if ag == "USE UPDATING AG":
            ag = universe.select_atoms(self.sel_acidic, updating=True)

        with pytest.raises(
            TypeError, match="must be either string or a static AtomGroup"
        ) as te:
            contacts.Contacts._get_atomgroup(universe, ag)

    def test_startframe(self, universe, client_Contacts):
        """test_startframe: TestContactAnalysis1: start frame set to 0 (resolution of
        Issue #624)

        """
        CA1 = self._run_Contacts(universe, client_Contacts=client_Contacts)
        assert len(CA1.results.timeseries) == universe.trajectory.n_frames

    def test_end_zero(self, universe, client_Contacts):
        """test_end_zero: TestContactAnalysis1: stop frame 0 is not ignored"""
        CA1 = self._run_Contacts(
            universe, client_Contacts=client_Contacts, stop=0
        )
        assert len(CA1.results.timeseries) == 0

    def test_slicing(self, universe, client_Contacts):
        start, stop, step = 10, 30, 5
        CA1 = self._run_Contacts(
            universe,
            client_Contacts=client_Contacts,
            start=start,
            stop=stop,
            step=step,
        )
        frames = np.arange(universe.trajectory.n_frames)[start:stop:step]
        assert len(CA1.results.timeseries) == len(frames)

    def test_villin_folded(self, client_Contacts):
        # one folded, one unfolded
        f = mda.Universe(contacts_villin_folded)
        u = mda.Universe(contacts_villin_unfolded)
        sel = "protein and not name H*"

        grF = f.select_atoms(sel)

        q = contacts.Contacts(
            u, select=(sel, sel), refgroup=(grF, grF), method="soft_cut"
        )
        q.run(**client_Contacts)

        results = soft_cut(f, u, sel, sel)
        assert_allclose(
            q.results.timeseries[:, 1], results[:, 1], rtol=0, atol=1.5e-7
        )

    def test_villin_unfolded(self, client_Contacts):
        # both folded
        f = mda.Universe(contacts_villin_folded)
        u = mda.Universe(contacts_villin_folded)
        sel = "protein and not name H*"

        grF = f.select_atoms(sel)

        q = contacts.Contacts(
            u, select=(sel, sel), refgroup=(grF, grF), method="soft_cut"
        )
        q.run(**client_Contacts)

        results = soft_cut(f, u, sel, sel)
        assert_allclose(
            q.results.timeseries[:, 1], results[:, 1], rtol=0, atol=1.5e-7
        )

    def test_hard_cut_method(self, universe, client_Contacts):
        ca = self._run_Contacts(universe, client_Contacts=client_Contacts)
        # fmt: off
        expected = [
            1., 0.58252427, 0.52427184, 0.55339806, 0.54368932,
            0.54368932, 0.51456311, 0.46601942, 0.48543689, 0.52427184,
            0.46601942, 0.58252427, 0.51456311, 0.48543689, 0.48543689,
            0.48543689, 0.46601942, 0.51456311, 0.49514563, 0.49514563,
            0.45631068, 0.47572816, 0.49514563, 0.50485437, 0.53398058,
            0.50485437, 0.51456311, 0.51456311, 0.49514563, 0.49514563,
            0.54368932, 0.50485437, 0.48543689, 0.55339806, 0.45631068,
            0.46601942, 0.53398058, 0.53398058, 0.46601942, 0.52427184,
            0.45631068, 0.46601942, 0.47572816, 0.46601942, 0.45631068,
            0.47572816, 0.45631068, 0.48543689, 0.4368932, 0.4368932,
            0.45631068, 0.50485437, 0.41747573, 0.4368932, 0.51456311,
            0.47572816, 0.46601942, 0.46601942, 0.47572816, 0.47572816,
            0.46601942, 0.45631068, 0.44660194, 0.47572816, 0.48543689,
            0.47572816, 0.42718447, 0.40776699, 0.37864078, 0.42718447,
            0.45631068, 0.4368932, 0.4368932, 0.45631068, 0.4368932,
            0.46601942, 0.45631068, 0.48543689, 0.44660194, 0.44660194,
            0.44660194, 0.42718447, 0.45631068, 0.44660194, 0.48543689,
            0.48543689, 0.44660194, 0.4368932, 0.40776699, 0.41747573,
            0.48543689, 0.45631068, 0.46601942, 0.47572816, 0.51456311,
            0.45631068, 0.37864078, 0.42718447,
        ]
        # fmt: on
        assert len(ca.results.timeseries) == len(expected)
        assert_allclose(
            ca.results.timeseries[:, 1], expected, rtol=0, atol=1.5e-7
        )

    def test_radius_cut_method(self, universe, client_Contacts):
        acidic = universe.select_atoms(self.sel_acidic)
        basic = universe.select_atoms(self.sel_basic)
        r = contacts.distance_array(acidic.positions, basic.positions)
        initial_contacts = contacts.contact_matrix(r, 6.0)
        expected = []
        for ts in universe.trajectory:
            r = contacts.distance_array(acidic.positions, basic.positions)
            expected.append(
                contacts.radius_cut_q(r[initial_contacts], None, radius=6.0)
            )

        ca = self._run_Contacts(
            universe, client_Contacts=client_Contacts, method="radius_cut"
        )
        assert_array_equal(ca.results.timeseries[:, 1], expected)

    @staticmethod
    def _is_any_closer(r, r0, dist=2.5):
        return np.any(r < dist)

    def test_own_method(self, universe, client_Contacts):
        ca = self._run_Contacts(
            universe,
            client_Contacts=client_Contacts,
            method=self._is_any_closer,
        )

        # fmt: off
        bound_expected = [
            1., 1., 0., 1., 1., 0., 0., 1., 0., 1., 1., 0., 0., 1., 0., 0.,
            0., 0., 1., 1., 0., 0., 0., 1., 0., 1., 0., 1., 1., 0., 1., 1.,
            1., 0., 0., 0., 0., 1., 0., 0., 1., 0., 1., 1., 1., 0., 1., 0.,
            0., 1., 1., 1., 0., 1., 0., 1., 1., 0., 0., 0., 1., 1., 1., 0.,
            0., 1., 0., 1., 1., 1., 1., 1., 1., 0., 1., 1., 0., 1., 0., 0.,
            1., 1., 0., 0., 1., 1., 1., 0., 1., 0., 0., 1., 0., 1., 1., 1.,
            1., 1.,
        ]
        # fmt: on
        assert_array_equal(ca.results.timeseries[:, 1], bound_expected)

    @staticmethod
    def _weird_own_method(r, r0):
        return "aaa"

    def test_own_method_no_array_cast(self, universe, client_Contacts):
        with pytest.raises(ValueError):
            self._run_Contacts(
                universe,
                client_Contacts=client_Contacts,
                method=self._weird_own_method,
                stop=2,
            )

    def test_non_callable_method(self, universe, client_Contacts):
        with pytest.raises(ValueError):
            self._run_Contacts(
                universe, client_Contacts=client_Contacts, method=2, stop=2
            )

    @pytest.mark.parametrize(
        "pbc,expected",
        [
            (
                True,
                [
                    1.0,
                    0.43138152,
                    0.3989021,
                    0.43824337,
                    0.41948765,
                    0.42223239,
                    0.41354071,
                    0.43641354,
                    0.41216834,
                    0.38334858,
                ],
            ),
            (
                False,
                [
                    1.0,
                    0.42327791,
                    0.39192399,
                    0.40950119,
                    0.40902613,
                    0.42470309,
                    0.41140143,
                    0.42897862,
                    0.41472684,
                    0.38574822,
                ],
            ),
        ],
    )
    def test_distance_box(self, pbc, expected, client_Contacts):
        u = mda.Universe(TPR, XTC)
        sel_basic = "(resname ARG LYS)"
        sel_acidic = "(resname ASP GLU)"
        acidic = u.select_atoms(sel_acidic)
        basic = u.select_atoms(sel_basic)

        r = contacts.Contacts(
            u,
            select=(sel_acidic, sel_basic),
            refgroup=(acidic, basic),
            radius=6.0,
            pbc=pbc,
        )
        r.run(**client_Contacts)
        assert_allclose(
            r.results.timeseries[:, 1], expected, rtol=0, atol=1.5e-7
        )

    def test_warn_deprecated_attr(self, universe, client_Contacts):
        """Test for warning message emitted on using deprecated `timeseries`
        attribute"""
        CA1 = self._run_Contacts(
            universe, client_Contacts=client_Contacts, stop=1
        )
        wmsg = "The `timeseries` attribute was deprecated in MDAnalysis"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(CA1.timeseries, CA1.results.timeseries)

    @pytest.mark.parametrize(
        "datafiles, expected", [((PSF, DCD), 0), ([TPR, XTC], 41814)]
    )
    def test_n_initial_contacts(self, datafiles, expected):
        """Test for n_initial_contacts attribute"""
        u = mda.Universe(*datafiles)
        select = ("protein", "not protein")
        refgroup = (u.select_atoms("protein"), u.select_atoms("not protein"))

        r = contacts.Contacts(u, select=select, refgroup=refgroup)
        assert_equal(r.n_initial_contacts, expected)


def test_q1q2(client_Contacts):
    u = mda.Universe(PSF, DCD)
    q1q2 = contacts.q1q2(u, "name CA", radius=8)
    q1q2.run(**client_Contacts)

    # fmt: off
    q1_expected = [
        1.0, 0.98092643, 0.97366031, 0.97275204, 0.97002725,
        0.97275204, 0.96276113, 0.96730245, 0.9582198, 0.96185286,
        0.95367847, 0.96276113, 0.9582198, 0.95186194, 0.95367847,
        0.95095368, 0.94187103, 0.95186194, 0.94277929, 0.94187103,
        0.9373297, 0.93642144, 0.93097184, 0.93914623, 0.93278837,
        0.93188011, 0.9373297, 0.93097184, 0.93188011, 0.92643052,
        0.92824705, 0.92915531, 0.92643052, 0.92461399, 0.92279746,
        0.92643052, 0.93278837, 0.93188011, 0.93369664, 0.9346049,
        0.9373297, 0.94096276, 0.9400545, 0.93642144, 0.9373297,
        0.9373297, 0.9400545, 0.93006358, 0.9400545, 0.93823797,
        0.93914623, 0.93278837, 0.93097184, 0.93097184, 0.92733878,
        0.92824705, 0.92279746, 0.92824705, 0.91825613, 0.92733878,
        0.92643052, 0.92733878, 0.93278837, 0.92733878, 0.92824705,
        0.93097184, 0.93278837, 0.93914623, 0.93097184, 0.9373297,
        0.92915531, 0.93188011, 0.93551317, 0.94096276, 0.93642144,
        0.93642144, 0.9346049, 0.93369664, 0.93369664, 0.93278837,
        0.93006358, 0.93278837, 0.93006358, 0.9346049, 0.92824705,
        0.93097184, 0.93006358, 0.93188011, 0.93278837, 0.93006358,
        0.92915531, 0.92824705, 0.92733878, 0.92643052, 0.93188011,
        0.93006358, 0.9346049, 0.93188011,
    ]
    # fmt: on
    assert_allclose(
        q1q2.results.timeseries[:, 1], q1_expected, rtol=0, atol=1.5e-7
    )

    # fmt: off
    q2_expected = [
        0.94649446, 0.94926199, 0.95295203, 0.95110701, 0.94833948,
        0.95479705, 0.94926199, 0.9501845, 0.94926199, 0.95387454,
        0.95202952, 0.95110701, 0.94649446, 0.94095941, 0.94649446,
        0.9400369, 0.94464945, 0.95202952, 0.94741697, 0.94649446,
        0.94188192, 0.94188192, 0.93911439, 0.94464945, 0.9400369,
        0.94095941, 0.94372694, 0.93726937, 0.93819188, 0.93357934,
        0.93726937, 0.93911439, 0.93911439, 0.93450185, 0.93357934,
        0.93265683, 0.93911439, 0.94372694, 0.93911439, 0.94649446,
        0.94833948, 0.95110701, 0.95110701, 0.95295203, 0.94926199,
        0.95110701, 0.94926199, 0.94741697, 0.95202952, 0.95202952,
        0.95202952, 0.94741697, 0.94741697, 0.94926199, 0.94280443,
        0.94741697, 0.94833948, 0.94833948, 0.9400369, 0.94649446,
        0.94741697, 0.94926199, 0.95295203, 0.94926199, 0.9501845,
        0.95664207, 0.95756458, 0.96309963, 0.95756458, 0.96217712,
        0.95756458, 0.96217712, 0.96586716, 0.96863469, 0.96494465,
        0.97232472, 0.97140221, 0.9695572, 0.97416974, 0.9695572,
        0.96217712, 0.96771218, 0.9704797, 0.96771218, 0.9695572,
        0.97140221, 0.97601476, 0.97693727, 0.98154982, 0.98431734,
        0.97601476, 0.9797048, 0.98154982, 0.98062731, 0.98431734,
        0.98616236, 0.9898524, 1.0,
    ]
    # fmt: on
    assert_allclose(
        q1q2.results.timeseries[:, 2], q2_expected, rtol=0, atol=1.5e-7
    )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""Analysis building blocks --- :mod:`MDAnalysis.analysis.base`
============================================================

MDAnalysis provides building blocks for creating analysis classes. One can
think of each analysis class as a "tool" that performs a specific analysis over
the trajectory frames and stores the results in the tool.

Analysis classes are derived from :class:`AnalysisBase` by subclassing. This
inheritance provides a common workflow and API for users and makes many
additional features automatically available (such as frame selections and a
verbose progressbar). The important points for analysis classes are:

#. Analysis tools are Python classes derived from :class:`AnalysisBase`.
#. When instantiating an analysis, the :class:`Universe` or :class:`AtomGroup`
   that the analysis operates on is provided together with any other parameters
   that are kept fixed for the specific analysis.
#. The analysis is performed with :meth:`~AnalysisBase.run` method. It has a
   common set of arguments such as being able to select the frames the analysis
   is performed on. The `verbose` keyword argument enables additional output. A
   progressbar is shown by default that also shows an estimate for the
   remaining time until the end of the analysis.
#. Results are always stored in the attribute :attr:`AnalysisBase.results`,
   which is an instance of :class:`Results`, a kind of dictionary that allows
   allows item access via attributes. Each analysis class decides what and how
   to store in :class:`Results` and needs to document it. For time series, the
   :attr:`AnalysisBase.times` contains the time stamps of the analyzed frames.


Example of using a standard analysis tool
-----------------------------------------

For example, the :class:`MDAnalysis.analysis.rms.RMSD` performs a
root-mean-square distance analysis in the following way:

.. code-block:: python

   import MDAnalysis as mda
   from MDAnalysisTests.datafiles import TPR, XTC

   from MDAnalysis.analysis import rms

   u = mda.Universe(TPR, XTC)

   # (2) instantiate analysis
   rmsd = rms.RMSD(u, select='name CA')

   # (3) the run() method can select frames in different ways
   # run on all frames (with progressbar)
   rmsd.run(verbose=True)

   # or start, stop, and step can be used
   rmsd.run(start=2, stop=8, step=2)

   # a list of frames to run the analysis on can be passed
   rmsd.run(frames=[0,2,3,6,9])

   # a list of booleans the same length of the trajectory can be used
   rmsd.run(frames=[True, False, True, True, False, False, True, False,
                    False, True])

   # (4) analyze the results, e.g., plot
   t = rmsd.times
   y = rmsd.results.rmsd[:, 2]   # RMSD at column index 2, see docs

   import matplotlib.pyplot as plt
   plt.plot(t, y)
   plt.xlabel("time (ps)")
   plt.ylabel("RMSD (Å)")


Writing new analysis tools
--------------------------

In order to write new analysis tools, derive a class from :class:`AnalysisBase`
and define at least the :meth:`_single_frame` method, as described in
:class:`AnalysisBase`.

.. SeeAlso::

   The chapter `Writing your own trajectory analysis`_ in the *User Guide*
   contains a step-by-step example for writing analysis tools with
   :class:`AnalysisBase`.

.. _`Writing your own trajectory analysis`:
   https://userguide.mdanalysis.org/stable/examples/analysis/custom_trajectory_analysis.html


If your analysis is operating independently on each frame, you might consider
making it **parallelizable** via adding a :meth:`get_supported_backends` method,
and appropriate aggregation function for each of its results. For example, if
you have your :meth:`_single_frame` method storing important values under
:attr:`self.results.timeseries`, you will write:

.. code-block:: python

    class MyAnalysis(AnalysisBase):
        _analysis_algorithm_is_parallelizable = True

        @classmethod
        def get_supported_backends(cls):
            return ('serial', 'multiprocessing', 'dask',)

        
        def _get_aggregator(self):
          return ResultsGroup(lookup={'timeseries': ResultsGroup.ndarray_vstack})

See :mod:`MDAnalysis.analysis.results` for more on aggregating results.

.. SeeAlso::

   :ref:`parallel-analysis`



Classes
-------

The :class:`MDAnalysis.results.Results` and :class:`AnalysisBase` classes
are the essential building blocks for almost all MDAnalysis tools in the
:mod:`MDAnalysis.analysis` module. They aim to be easily useable and
extendable.

:class:`AnalysisFromFunction` and the :func:`analysis_class` functions are
simple wrappers that make it even easier to create fully-featured analysis
tools if only the single-frame analysis function needs to be written.

"""
import inspect
import itertools
import logging
import warnings
from functools import partial
from typing import Iterable, Union

import numpy as np
from .. import coordinates
from ..core.groups import AtomGroup
from ..lib.log import ProgressBar

from .backends import (
    BackendDask,
    BackendMultiprocessing,
    BackendSerial,
    BackendBase,
)
from .results import Results, ResultsGroup

logger = logging.getLogger(__name__)


class AnalysisBase(object):
    r"""Base class for defining multi-frame analysis

    The class is designed as a template for creating multi-frame analyses.
    This class will automatically take care of setting up the trajectory
    reader for iterating, and it offers to show a progress meter.
    Computed results are stored inside the :attr:`results` attribute.

    To define a new Analysis, :class:`AnalysisBase` needs to be subclassed
    and :meth:`_single_frame` must be defined. It is also possible to define
    :meth:`_prepare` and :meth:`_conclude` for pre- and post-processing.
    All results should be stored as attributes of the
    :class:`MDAnalysis.analysis.results.Results` container.

    .. Note::
       The instance attributes are created during and on conclusion of
       calling the :meth:`AnalysisBase.run` method. Accessing an attribute
       before it has been created will raise an :exc:`AttributeError`.


    Parameters
    ----------
    trajectory : MDAnalysis.coordinates.base.ReaderBase
        A trajectory Reader
    verbose : bool, optional
        Turn on more logging and debugging

    Attributes
    ----------
    times: numpy.ndarray
        Array of times of the Timesteps that were analyzed.
        Only exists after calling :meth:`AnalysisBase.run`.
    frames: numpy.ndarray
        Array of frame indices that were analyzed.
        Only exists after calling :meth:`AnalysisBase.run`.
    results: :class:`Results`
        Results of calculation are stored here, after call
        to :meth:`AnalysisBase.run`.
    n_frames: int
        number of *analyzed* frames, i.e., after taking into account
        the `start`, `stop`, and `step` values from
        :meth:`AnalysisBase.run`.
        Only exists after calling :meth:`AnalysisBase.run`.
    start: int
        Frame index of the first trajectory frame that was analyzed.
        Only exists after calling :meth:`AnalysisBase.run`.
    stop: int
        Frame index of the last trajectory frame that was analyzed.
        Only exists after calling :meth:`AnalysisBase.run`.
    step: int
        Every `step` frame was analyzed, as ``trajectory[start:stop:step]``.
        Only exists after calling :meth:`AnalysisBase.run`.


    Example
    -------
    .. code-block:: python

       from MDAnalysis.analysis.base import AnalysisBase

       class NewAnalysis(AnalysisBase):
           def __init__(self, atomgroup, parameter, **kwargs):
               super(NewAnalysis, self).__init__(atomgroup.universe.trajectory,
                                                 **kwargs)
               self._parameter = parameter
               self._ag = atomgroup

           def _prepare(self):
               # OPTIONAL
               # Called before iteration on the trajectory has begun.
               # Data structures can be set up at this time
               self.results.example_result = []

           def _single_frame(self):
               # REQUIRED
               # Called after the trajectory is moved onto each new frame.
               # store an example_result of `some_function` for a single frame
               self.results.example_result.append(some_function(self._ag,
                                                                self._parameter))

           def _conclude(self):
               # OPTIONAL
               # Called once iteration on the trajectory is finished.
               # Apply normalisation and averaging to results here.
               self.results.example_result = np.asarray(self.example_result)
               self.results.example_result /=  np.sum(self.result)

    Afterwards the new analysis can be run like this

    .. code-block:: python

       import MDAnalysis as mda
       from MDAnalysisTests.datafiles import PSF, DCD

       u = mda.Universe(PSF, DCD)

       na = NewAnalysis(u.select_atoms('name CA'), 35)
       na.run(start=10, stop=20)
       print(na.results.example_result)
       # results can also be accessed by key
       print(na.results["example_result"])


    .. versionchanged:: 1.0.0
        Support for setting `start`, `stop`, and `step` has been removed. These
        should now be directly passed to :meth:`AnalysisBase.run`.

    .. versionchanged:: 2.0.0
        Added :attr:`results`

    .. versionchanged:: 2.8.0
        Added ability to run analysis in parallel using either a
        built-in backend (`multiprocessing` or `dask`) or a custom
        `backends.BackendBase` instance with an implemented `apply` method
        that is used to run the computations.
    """

    @classmethod
    def get_supported_backends(cls):
        """Tuple with backends supported by the core library for a given class.
        User can pass either one of these values as ``backend=...`` to
        :meth:`run()` method, or a custom object that has ``apply`` method
        (see documentation for :meth:`run()`):

         - 'serial': no parallelization
         - 'multiprocessing': parallelization using `multiprocessing.Pool`
         - 'dask': parallelization using `dask.delayed.compute()`. Requires
           installation of `mdanalysis[dask]`

        If you want to add your own backend to an existing class, pass a
        :class:`backends.BackendBase` subclass (see its documentation to learn
        how to implement it properly), and specify ``unsupported_backend=True``.

        Returns
        -------
        tuple
            names of built-in backends that can be used in :meth:`run(backend=...)`


        .. versionadded:: 2.8.0
        """
        return ("serial",)

    # class authors: override _analysis_algorithm_is_parallelizable
    # in derived classes and only set to True if you have confirmed
    # that your algorithm works reliably when parallelized with
    # the split-apply-combine approach (see docs)
    _analysis_algorithm_is_parallelizable = False

    @property
    def parallelizable(self):
        """Boolean mark showing that a given class can be parallelizable with
        split-apply-combine procedure. Namely, if we can safely distribute
        :meth:`_single_frame` to multiple workers and then combine them with a
        proper :meth:`_conclude` call. If set to ``False``, no backends except
        for ``serial`` are supported.

        .. note::   If you want to check parallelizability of the whole class, without
                    explicitly creating an instance of the class, see
                    :attr:`_analysis_algorithm_is_parallelizable`. Note that you
                    setting it to other value will break things if the algorithm
                    behind the analysis is not trivially parallelizable.


        Returns
        -------
        bool
            if a given ``AnalysisBase`` subclass instance
            is parallelizable with split-apply-combine, or not


        .. versionadded:: 2.8.0
        """
        return self._analysis_algorithm_is_parallelizable

    def __init__(self, trajectory, verbose=False, **kwargs):
        self._trajectory = trajectory
        self._verbose = verbose
        self.results = Results()

    def _define_run_frames(
        self, trajectory, start=None, stop=None, step=None, frames=None
    ) -> Union[slice, np.ndarray]:
        """Defines limits for the whole run, as passed by self.run() arguments

        Parameters
        ----------
        trajectory : mda.Reader
            a trajectory Reader
        start : int, optional
            start frame of analysis, by default None
        stop : int, optional
            stop frame of analysis, by default None
        step : int, optional
            number of frames to skip between each analysed frame, by default None
        frames : array_like, optional
            array of integers or booleans to slice trajectory; cannot be
            combined with ``start``, ``stop``, ``step``; by default None

        Returns
        -------
        Union[slice, np.ndarray]
            Appropriate slicer for the trajectory that would give correct iteraction
            order via trajectory[slicer]

        Raises
        ------
        ValueError
            if *both* `frames` and at least one of ``start``, ``stop``, or ``step``
            is provided (i.e. set to not ``None`` value).


        .. versionadded:: 2.8.0
        """
        self._trajectory = trajectory
        if frames is not None:
            if not all(opt is None for opt in [start, stop, step]):
                raise ValueError(
                    "start/stop/step cannot be combined with frames"
                )
            slicer = frames
        else:
            start, stop, step = trajectory.check_slice_indices(
                start, stop, step
            )
            slicer = slice(start, stop, step)
        self.start, self.stop, self.step = start, stop, step
        return slicer

    def _prepare_sliced_trajectory(self, slicer: Union[slice, np.ndarray]):
        """Prepares sliced trajectory for use in subsequent parallel computations:
        namely, assigns self._sliced_trajectory and its appropriate attributes,
        self.n_frames, self.frames and self.times.

        Parameters
        ----------
        slicer : Union[slice, np.ndarray]
            appropriate slicer for the trajectory


        .. versionadded:: 2.8.0
        """
        self._sliced_trajectory = self._trajectory[slicer]
        self.n_frames = len(self._sliced_trajectory)
        self.frames = np.zeros(self.n_frames, dtype=int)
        self.times = np.zeros(self.n_frames)

    def _setup_frames(
        self, trajectory, start=None, stop=None, step=None, frames=None
    ):
        """Pass a Reader object and define the desired iteration pattern
        through the trajectory

        Parameters
        ----------
        trajectory : mda.Reader
            A trajectory Reader
        start : int, optional
            start frame of analysis
        stop : int, optional
            stop frame of analysis
        step : int, optional
            number of frames to skip between each analysed frame
        frames : array_like, optional
            array of integers or booleans to slice trajectory; cannot be
            combined with ``start``, ``stop``, ``step``

            .. versionadded:: 2.2.0

        Raises
        ------
        ValueError
            if *both* `frames` and at least one of ``start``, ``stop``, or
            ``frames`` is provided (i.e., set to another value than ``None``)


        .. versionchanged:: 1.0.0
            Added .frames and .times arrays as attributes

        .. versionchanged:: 2.2.0
            Added ability to iterate through trajectory by passing a list of
            frame indices in the `frames` keyword argument

        .. versionchanged:: 2.8.0
            Split function into two: :meth:`_define_run_frames` and
            :meth:`_prepare_sliced_trajectory`: first one defines the limits
            for the whole run and is executed once during :meth:`run` in
            :meth:`_setup_frames`, second one prepares sliced trajectory for
            each of the workers and gets executed twice: one time in
            :meth:`_setup_frames` for the whole trajectory, second time in
            :meth:`_compute` for each of the computation groups.
        """
        slicer = self._define_run_frames(trajectory, start, stop, step, frames)
        self._prepare_sliced_trajectory(slicer)

    def _single_frame(self):
        """Calculate data from a single frame of trajectory

        Don't worry about normalising, just deal with a single frame.
        Attributes accessible during your calculations:

          - ``self._frame_index``: index of the frame in results array
          - ``self._ts`` -- Timestep instance
          - ``self._sliced_trajectory`` -- trajectory that you're iterating over
          - ``self.results`` -- :class:`MDAnalysis.analysis.results.Results` instance
            holding run results initialized in :meth:`_prepare`.
        """
        raise NotImplementedError("Only implemented in child classes")

    def _prepare(self):
        """
        Set things up before the analysis loop begins.

        Notes
        -----
        ``self.results`` is initialized already in :meth:`self.__init__` with an
        empty instance of :class:`MDAnalysis.analysis.results.Results` object.
        You can still call your attributes as if they were usual ones,
        ``Results`` just keeps track of that to be able to run a proper
        aggregation after a parallel run, if necessary.
        """
        pass  # pylint: disable=unnecessary-pass

    def _conclude(self):
        """Finalize the results you've gathered.

        Called at the end of the :meth:`run` method to finish everything up.

        Notes
        -----
        Aggregation of results from individual workers happens in
        :meth:`self.run()`, so here you have to implement everything as if you
        had a non-parallel run. If you want to enable proper aggregation for
        parallel runs for you analysis class, implement ``self._get_aggregator``
        and check :mod:`MDAnalysis.analysis.results` for how to use it.
        """
        pass  # pylint: disable=unnecessary-pass

    def _compute(
        self,
        indexed_frames: np.ndarray,
        verbose: bool = None,
        *,
        progressbar_kwargs=None,
    ) -> "AnalysisBase":
        """Perform the calculation on a balanced slice of frames
        that have been setup prior to that using _setup_computation_groups()

        Parameters
        ----------
        indexed_frames : np.ndarray
            np.ndarray of (n, 2) shape, where first column is frame iteration
            indices and second is frame numbers

        verbose : bool, optional
            Turn on verbosity

        progressbar_kwargs : dict, optional
            ProgressBar keywords with custom parameters regarding progress bar
            position, etc; see :class:`MDAnalysis.lib.log.ProgressBar`
            for full list.


        .. versionadded:: 2.8.0
        """
        if progressbar_kwargs is None:
            progressbar_kwargs = {}
        logger.info("Choosing frames to analyze")
        # if verbose unchanged, use class default
        verbose = (
            getattr(self, "_verbose", False) if verbose is None else verbose
        )

        frames = indexed_frames[:, 1]

        logger.info("Starting preparation")
        self._prepare_sliced_trajectory(slicer=frames)
        self._prepare()
        if len(frames) == 0:  # if `frames` were empty in `run` or `stop=0`
            return self

        for idx, ts in enumerate(
            ProgressBar(
                self._sliced_trajectory, verbose=verbose, **progressbar_kwargs
            )
        ):
            self._frame_index = idx  # accessed later by subclasses
            self._ts = ts
            self.frames[idx] = ts.frame
            self.times[idx] = ts.time
            self._single_frame()
        logger.info("Finishing up")
        return self

    def _setup_computation_groups(
        self,
        n_parts: int,
        start: int = None,
        stop: int = None,
        step: int = None,
        frames: Union[slice, np.ndarray] = None,
    ) -> list[np.ndarray]:
        """
        Splits the trajectory frames, defined by ``start/stop/step`` or
        ``frames``, into ``n_parts`` even groups, preserving their indices.

        Parameters
        ----------
        n_parts : int
            number of parts to split the workload into
        start : int, optional
            start frame
        stop : int, optional
            stop frame
        step : int, optional
            step size for analysis (1 means to read every frame)
        frames : array_like, optional
            array of integers or booleans to slice trajectory; ``frames`` can
            only be used *instead* of ``start``, ``stop``, and ``step``. Setting
            *both* ``frames`` and at least one of ``start``, ``stop``, ``step``
            to a non-default value will raise a :exc:`ValueError`.

        Raises
        ------
        ValueError
            if *both* ``frames`` and at least one of ``start``, ``stop``, or
            ``frames`` is provided (i.e., set to another value than ``None``)

        Returns
        -------
        computation_groups : list[np.ndarray]
            list of (n, 2) shaped np.ndarrays with frame indices and numbers


        .. versionadded:: 2.8.0
        """
        if frames is None:
            start, stop, step = self._trajectory.check_slice_indices(
                start, stop, step
            )
            used_frames = np.arange(start, stop, step)
        elif not all(opt is None for opt in [start, stop, step]):
            raise ValueError("start/stop/step cannot be combined with frames")
        else:
            used_frames = frames

        if all(isinstance(obj, bool) for obj in used_frames):
            arange = np.arange(len(used_frames))
            used_frames = arange[used_frames]

        # similar to list(enumerate(frames))
        enumerated_frames = np.vstack(
            [np.arange(len(used_frames)), used_frames]
        ).T
        if len(enumerated_frames) == 0:
            return [np.empty((0, 2), dtype=np.int64)]
        elif len(enumerated_frames) < n_parts:
            # Issue #4685
            n_parts = len(enumerated_frames)
            warnings.warn(
                f"Set `n_parts` to {n_parts} to match the total "
                "number of frames being analyzed"
            )

        return np.array_split(enumerated_frames, n_parts)

    def _configure_backend(
        self,
        backend: Union[str, BackendBase],
        n_workers: int,
        unsupported_backend: bool = False,
    ) -> BackendBase:
        """Matches a passed backend string value with class attributes
        :attr:`parallelizable` and :meth:`get_supported_backends()`
        to check if downstream calculations can be performed.

        Parameters
        ----------
        backend : Union[str, BackendBase]
            backend to be used:
               - ``str`` is matched to a builtin backend (one of "serial",
                 "multiprocessing" and "dask")
               - ``BackendBase`` subclass is checked for the presence of
                 an :meth:`apply` method
        n_workers : int
            positive integer with number of workers (processes, in case of
            built-in backends) to split the work between
        unsupported_backend : bool, optional
            if you want to run your custom backend on a parallelizable class
            that has not been tested by developers, by default ``False``

        Returns
        -------
        BackendBase
            instance of a ``BackendBase`` class that will be used for computations

        Raises
        ------
        ValueError
            if :attr:`parallelizable` is set to ``False`` but backend is
            not ``serial``
        ValueError
            if :attr:`parallelizable` is ``True`` and custom backend instance is used
            without specifying ``unsupported_backend=True``
        ValueError
            if your trajectory has associated parallelizable transformations
            but backend is not serial
        ValueError
            if ``n_workers`` was specified twice -- in the run() method and durin
            ``__init__`` of a custom backend
        ValueError
            if your backend object instance doesn't have an ``apply`` method


        .. versionadded:: 2.8.0
        """
        builtin_backends = {
            "serial": BackendSerial,
            "multiprocessing": BackendMultiprocessing,
            "dask": BackendDask,
        }

        backend_class = builtin_backends.get(backend, backend)
        supported_backend_classes = [
            builtin_backends.get(b) for b in self.get_supported_backends()
        ]

        # check for serial-only classes
        if not self.parallelizable and backend_class is not BackendSerial:
            raise ValueError(f"Can not parallelize class {self.__class__}")

        # make sure user enabled 'unsupported_backend=True' for custom classes
        if (
            not unsupported_backend
            and self.parallelizable
            and backend_class not in supported_backend_classes
        ):
            raise ValueError(
                (
                    f"Must specify 'unsupported_backend=True'"
                    f"if you want to use a custom {backend_class=} for {self.__class__}"
                )
            )

        # check for the presence of parallelizable transformations
        if backend_class is not BackendSerial and any(
            not t.parallelizable for t in self._trajectory.transformations
        ):
            raise ValueError(
                (
                    "Trajectory should not have "
                    "associated unparallelizable transformations"
                )
            )

        # conclude mapping from string to backend class if it's a builtin backend
        if isinstance(backend, str):
            return backend_class(n_workers=n_workers)

        # make sure we haven't specified n_workers twice
        if (
            isinstance(backend, BackendBase)
            and n_workers is not None
            and hasattr(backend, "n_workers")
            and backend.n_workers != n_workers
        ):
            raise ValueError(
                (
                    f"n_workers specified twice: in {backend.n_workers=}"
                    f"and in run({n_workers=}). Remove it from run()"
                )
            )

        # or pass along an instance of the class itself
        # after ensuring it has apply method
        if not isinstance(backend, BackendBase) or not hasattr(
            backend, "apply"
        ):
            raise ValueError(
                (
                    f"{backend=} is invalid: should have 'apply' method "
                    "and be instance of MDAnalysis.analysis.backends.BackendBase"
                )
            )
        return backend

    def run(
        self,
        start: int = None,
        stop: int = None,
        step: int = None,
        frames: Iterable = None,
        verbose: bool = None,
        n_workers: int = None,
        n_parts: int = None,
        backend: Union[str, BackendBase] = None,
        *,
        unsupported_backend: bool = False,
        progressbar_kwargs=None,
    ):
        """Perform the calculation

        Parameters
        ----------
        start : int, optional
            start frame of analysis
        stop : int, optional
            stop frame of analysis
        step : int, optional
            number of frames to skip between each analysed frame
        frames : array_like, optional
            array of integers or booleans to slice trajectory; ``frames`` can
            only be used *instead* of ``start``, ``stop``, and ``step``. Setting
            *both* ``frames`` and at least one of ``start``, ``stop``, ``step``
            to a non-default value will raise a :exc:`ValueError`.

            .. versionadded:: 2.2.0
        verbose : bool, optional
            Turn on verbosity

        progressbar_kwargs : dict, optional
            ProgressBar keywords with custom parameters regarding progress bar
            position, etc; see :class:`MDAnalysis.lib.log.ProgressBar`
            for full list. Available only for ``backend='serial'``
        backend : Union[str, BackendBase], optional
            By default, performs calculations in a serial fashion.
            Otherwise, user can choose a backend: ``str`` is matched to a
            builtin backend (one of ``serial``, ``multiprocessing`` and
            ``dask``), or a :class:`MDAnalysis.analysis.results.BackendBase`
            subclass.

            .. versionadded:: 2.8.0
        n_workers : int
            positive integer with number of workers (processes, in case of
            built-in backends) to split the work between

            .. versionadded:: 2.8.0
        n_parts : int, optional
            number of parts to split computations across. Can be more than
            number of workers.

            .. versionadded:: 2.8.0
        unsupported_backend : bool, optional
            if you want to run your custom backend on a parallelizable class
            that has not been tested by developers, by default False

            .. versionadded:: 2.8.0


        .. versionchanged:: 2.2.0
            Added ability to analyze arbitrary frames by passing a list of
            frame indices in the `frames` keyword argument.

        .. versionchanged:: 2.5.0
            Add `progressbar_kwargs` parameter,
            allowing to modify description, position etc of tqdm progressbars

        .. versionchanged:: 2.8.0
            Introduced ``backend``, ``n_workers``, ``n_parts`` and
            ``unsupported_backend`` keywords, and refactored the method logic to
            support parallelizable execution.
        """
        # default to serial execution
        backend = "serial" if backend is None else backend

        progressbar_kwargs = (
            {} if progressbar_kwargs is None else progressbar_kwargs
        )
        if (progressbar_kwargs or verbose) and not (
            backend == "serial" or isinstance(backend, BackendSerial)
        ):
            raise ValueError(
                "Can not display progressbar with non-serial backend"
            )

        # if number of workers not specified, try getting the number from
        # the backend instance if possible, or set to 1
        if n_workers is None:
            n_workers = (
                backend.n_workers
                if isinstance(backend, BackendBase)
                and hasattr(backend, "n_workers")
                else 1
            )

        # set n_parts and check that is has a reasonable value
        n_parts = n_workers if n_parts is None else n_parts

        # do this as early as possible to check client parameters
        # before any computations occur
        executor = self._configure_backend(
            backend=backend,
            n_workers=n_workers,
            unsupported_backend=unsupported_backend,
        )
        if (
            hasattr(executor, "n_workers") and n_parts < executor.n_workers
        ):  # using executor's value here for non-default executors
            warnings.warn(
                (
                    f"Analysis not making use of all workers: "
                    f"{executor.n_workers=} is greater than {n_parts=}"
                )
            )

        # start preparing the run
        worker_func = partial(
            self._compute,
            progressbar_kwargs=progressbar_kwargs,
            verbose=verbose,
        )
        self._setup_frames(
            trajectory=self._trajectory,
            start=start,
            stop=stop,
            step=step,
            frames=frames,
        )
        computation_groups = self._setup_computation_groups(
            start=start, stop=stop, step=step, frames=frames, n_parts=n_parts
        )

        # get all results from workers in other processes.
        # we need `AnalysisBase` classes
        # since they hold `frames`, `times` and `results` attributes
        remote_objects: list["AnalysisBase"] = executor.apply(
            worker_func, computation_groups
        )
        self.frames = np.hstack([obj.frames for obj in remote_objects])
        self.times = np.hstack([obj.times for obj in remote_objects])

        # aggregate results from results obtained in remote workers
        remote_results = [obj.results for obj in remote_objects]
        results_aggregator = self._get_aggregator()
        self.results = results_aggregator.merge(remote_results)

        self._conclude()
        return self

    def _get_aggregator(self) -> ResultsGroup:
        """Returns a default aggregator that takes entire results
        if there is a single object, and raises ValueError otherwise

        Returns
        -------
        ResultsGroup
            aggregating object


        .. versionadded:: 2.8.0
        """
        return ResultsGroup(lookup=None)


class AnalysisFromFunction(AnalysisBase):
    r"""Create an :class:`AnalysisBase` from a function working on AtomGroups

    Parameters
    ----------
    function : callable
        function to evaluate at each frame
    trajectory : MDAnalysis.coordinates.Reader, optional
        trajectory to iterate over. If ``None`` the first AtomGroup found in
        args and kwargs is used as a source for the trajectory.
    *args : list
        arguments for `function`
    **kwargs : dict
        arguments for `function` and :class:`AnalysisBase`

    Attributes
    ----------
    results.frames : numpy.ndarray
            simulation frames used in analysis
    results.times : numpy.ndarray
            simulation times used in analysis
    results.timeseries : numpy.ndarray
            Results for each frame of the wrapped function,
            stored after call to :meth:`AnalysisFromFunction.run`.

    Raises
    ------
    ValueError
        if `function` has the same `kwargs` as :class:`AnalysisBase`

    Example
    -------
    .. code-block:: python

        def rotation_matrix(mobile, ref):
            return mda.analysis.align.rotation_matrix(mobile, ref)[0]

        rot = AnalysisFromFunction(rotation_matrix, trajectory,
                                    mobile, ref).run()
        print(rot.results.timeseries)


    .. versionchanged:: 1.0.0
        Support for directly passing the `start`, `stop`, and `step` arguments
        has been removed. These should instead be passed to
        :meth:`AnalysisFromFunction.run`.

    .. versionchanged:: 2.0.0
        Former :attr:`results` are now stored as :attr:`results.timeseries`

    .. versionchanged:: 2.8.0
        Added :meth:`get_supported_backends()`, introducing 'serial', 'multiprocessing'
        and 'dask' backends.
    """

    _analysis_algorithm_is_parallelizable = True

    @classmethod
    def get_supported_backends(cls):
        return ("serial", "multiprocessing", "dask")

    def __init__(self, function, trajectory=None, *args, **kwargs):
        if (trajectory is not None) and (
            not isinstance(trajectory, coordinates.base.ProtoReader)
        ):
            args = (trajectory,) + args
            trajectory = None

        if trajectory is None:
            # all possible places to find trajectory
            for arg in itertools.chain(args, kwargs.values()):
                if isinstance(arg, AtomGroup):
                    trajectory = arg.universe.trajectory
                    break

        if trajectory is None:
            raise ValueError("Couldn't find a trajectory")

        self.function = function
        self.args = args

        self.kwargs = kwargs

        super(AnalysisFromFunction, self).__init__(trajectory)

    def _prepare(self):
        self.results.timeseries = []

    def _get_aggregator(self):
        return ResultsGroup({"timeseries": ResultsGroup.flatten_sequence})

    def _single_frame(self):
        self.results.timeseries.append(
            self.function(*self.args, **self.kwargs)
        )

    def _conclude(self):
        self.results.frames = self.frames
        self.results.times = self.times
        self.results.timeseries = np.asarray(self.results.timeseries)


def analysis_class(function):
    r"""Transform a function operating on a single frame to an
    :class:`AnalysisBase` class.

    Parameters
    ----------
    function : callable
        function to evaluate at each frame

    Attributes
    ----------
    results.frames : numpy.ndarray
            simulation frames used in analysis
    results.times : numpy.ndarray
            simulation times used in analysis
    results.timeseries : numpy.ndarray
            Results for each frame of the wrapped function,
            stored after call to :meth:`AnalysisFromFunction.run`.

    Raises
    ------
    ValueError
        if `function` has the same `kwargs` as :class:`AnalysisBase`

    Examples
    --------

    For use in a library, we recommend the following style

    .. code-block:: python

        def rotation_matrix(mobile, ref):
            return mda.analysis.align.rotation_matrix(mobile, ref)[0]
        RotationMatrix = analysis_class(rotation_matrix)

    It can also be used as a decorator

    .. code-block:: python

        @analysis_class
        def RotationMatrix(mobile, ref):
            return mda.analysis.align.rotation_matrix(mobile, ref)[0]

        rot = RotationMatrix(u.trajectory, mobile, ref).run(step=2)
        print(rot.results.timeseries)


    .. versionchanged:: 2.0.0
        Former :attr:`results` are now stored as :attr:`results.timeseries`
    """

    class WrapperClass(AnalysisFromFunction):
        def __init__(self, trajectory=None, *args, **kwargs):
            super(WrapperClass, self).__init__(
                function, trajectory, *args, **kwargs
            )

        @classmethod
        def get_supported_backends(cls):
            return ("serial", "dask")

    return WrapperClass


def _filter_baseanalysis_kwargs(function, kwargs):
    """
    Create two dictionaries with `kwargs` separated for `function` and
    :class:`AnalysisBase`

    Parameters
    ----------
    function : callable
        function to be called
    kwargs : dict
        keyword argument dictionary

    Returns
    -------
    base_args : dict
        dictionary of AnalysisBase kwargs
    kwargs : dict
        kwargs without AnalysisBase kwargs

    Raises
    ------
    ValueError
        if `function` has the same `kwargs` as :class:`AnalysisBase`

    """
    try:
        # pylint: disable=deprecated-method
        base_argspec = inspect.getfullargspec(AnalysisBase.__init__)
    except AttributeError:
        # pylint: disable=deprecated-method
        base_argspec = inspect.getargspec(AnalysisBase.__init__)

    n_base_defaults = len(base_argspec.defaults)
    base_kwargs = {
        name: val
        for name, val in zip(
            base_argspec.args[-n_base_defaults:], base_argspec.defaults
        )
    }

    try:
        # pylint: disable=deprecated-method
        argspec = inspect.getfullargspec(function)
    except AttributeError:
        # pylint: disable=deprecated-method
        argspec = inspect.getargspec(function)

    for base_kw in base_kwargs.keys():
        if base_kw in argspec.args:
            raise ValueError(
                "argument name '{}' clashes with AnalysisBase argument."
                "Now allowed are: {}".format(base_kw, base_kwargs.keys())
            )

    base_args = {}
    for argname, default in base_kwargs.items():
        base_args[argname] = kwargs.pop(argname, default)

    return base_args, kwargs



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
Location of data files for the MDAnalysis unit tests
====================================================

Real MD simulation data, used for examples and the unit tests::

  from MDAnalysis.tests.datafiles import *

.. Note::

   The actual files are contained in the separate
   :mod:`MDAnalysisTests` package which must be downloaded from
   http://pypi.python.org/pypi/MDAnalysisTests and installed.
"""

try:
    from MDAnalysisTests.datafiles import *
except ImportError:
    print("*** ERROR ***")
    print("In order to run the MDAnalysis test cases you must install the")
    print("MDAnalysisTestData package (which has been separated from the ")
    print("library code itself since release 0.7.4). Go to ")
    print()
    print("     http://pypi.python.org/pypi/MDAnalysisTests")
    print()
    print("and download and install the `MDAnalysisTests-x.y.z.tar.gz'")
    print("that matches your MDAnalysis release.")
    raise ImportError("MDAnalysisTests package not installed.") from None



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""MMTF trajectory reader --- :mod:`MDAnalysis.coordinates.MMTF`
================================================================

Reads coordinates data from the Macromolecular Transmission Format format
(MMTF_). This should generally be a quicker alternative to PDB.

.. versionadded:: 0.16.0
.. versionchanged:: 0.20.0
   Unit cell is now properly treated as optional

Classes
-------

.. autoclass:: MMTFReader
   :members:

.. _MMTF: https://www.rcsb.org/news/feature/65a1af31c76ca3abcc925d0c

"""
import warnings
import mmtf

from . import base
from ..core.universe import Universe
from ..lib.util import cached, store_init_arguments
from ..due import due, Doi


def _parse_mmtf(fn):
    if fn.endswith("gz"):
        return mmtf.parse_gzip(fn)
    else:
        return mmtf.parse(fn)


class MMTFReader(base.SingleFrameReaderBase):
    """Coordinate reader for the Macromolecular Transmission Format format (MMTF_).


    .. deprecated:: 2.8.0
       The MMTF format is no longer supported / serviced by the
       Protein Data Bank. The Reader will be removed in version 3.0.
       Users are encouraged to instead use alternative PDB formats.
    """

    format = "MMTF"

    @store_init_arguments
    def __init__(self, filename, convert_units=True, n_atoms=None, **kwargs):
        wmsg = (
            "The MMTF Reader is deprecated and will be removed in "
            "MDAnalysis version 3.0.0"
        )
        warnings.warn(wmsg, DeprecationWarning)

        super(MMTFReader, self).__init__(
            filename, convert_units, n_atoms, **kwargs
        )

    @staticmethod
    def _format_hint(thing):
        """Can this Reader read *thing*?

        .. versionadded:: 1.0.0
        """
        return isinstance(thing, mmtf.MMTFDecoder)

    @due.dcite(
        Doi("10.1371/journal.pcbi.1005575"),
        description="MMTF Reader",
        path="MDAnalysis.coordinates.MMTF",
    )
    def _read_first_frame(self):
        # TOOD: Check units?
        if isinstance(self.filename, mmtf.MMTFDecoder):
            top = self.filename
        else:
            top = _parse_mmtf(self.filename)
        self.n_atoms = top.num_atoms

        self.ts = ts = self._Timestep(self.n_atoms, **self._ts_kwargs)
        ts._pos[:, 0] = top.x_coord_list
        ts._pos[:, 1] = top.y_coord_list
        ts._pos[:, 2] = top.z_coord_list
        if not top.unit_cell is None:
            # optional field
            ts.dimensions = top.unit_cell

        return ts



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
MDAnalysis topology tables
==========================

The module contains static lookup tables for atom typing etc. The
tables are dictionaries that are indexed by the element.

.. autodata:: atomelements
.. autodata:: masses
.. autodata:: vdwradii

The original raw data are stored as multi-line strings that are
translated into dictionaries with :func:`kv2dict`. In the future,
these tables might be moved into external data files; see
:func:`kv2dict` for explanation of the file format.

.. autofunction:: kv2dict

The raw tables are stored in the strings

.. autodata:: TABLE_ATOMELEMENTS
.. autodata:: TABLE_MASSES
.. autodata:: TABLE_VDWRADII
"""

from typing import Any


def kv2dict(s, convertor: Any = str):
    """Primitive ad-hoc parser of a key-value record list.

    * The string *s* should contain each key-value pair on a separate
      line (separated by newline). The first white space after the key
      separates key and value.

    * Empty lines are allowed.

    * Comment lines (starting with #) are allowed.

    * Leading whitespace is ignored.

    The *convertor* is a function that converts its single argument to
    a valid Python type. The default is :func:`str` but other
    possibilities are :func:`int` (for integers) or :func:`float` for
    floating point numbers.
    """
    d = {}
    lines = s.splitlines()
    for line in lines:
        line = line.lstrip()
        values = line.split(None, 1)
        if len(values) == 0 or line.startswith("#"):
            continue
        d[values[0]] = convertor(values[1])
    return d


#: Table with hard-coded special atom names, used for guessing atom types
#: with :func:`MDAnalysis.topology.core.guess_atom_element`.
TABLE_ATOMELEMENTS = """
# translation of atomnames to types/element
# based on CHARMM and AMBER usage with a little bit of GROMOS (and PROPKA)
# NOTE: CL might be ambiguous and is interpreted as chloride!

# --------- ------------------
# atomname   element
# --------- ------------------

# Bromide
BR           BR

# Calcium
CAL          CA
C0           CA
CA2+         CA

# Cesium
CES          CS

# Chloride
CLA          CL
CLAL         CL
CL           CL
CL-          CL

# Iodide
IOD          I

# Iron
FE           FE
FE2          FE

# Lithium
LIT          LI
LI           LI
LI+          LI
QL           LI

# Magnesium
MG           MG
MG2+         MG

# Noble gases
## XXX collides with NE, HE in Arg  XXX
## XXX so we remove the noble gases XXX
##HE           HE
##NE           NE

# Potassium
K            K
POT          K
K+           K
QK           K

# Sodium
SOD          NA
NA           NA
NA+          NA
QN           NA

# Zink
ZN           ZN

# Copper
CU           CU

# Cesium
CS           CS
CS+          CS
CES          CS

# Cerium??
QC           CE

# Rubidium
RB           RB
QR           RB

# special carbons (Amber?)
BC           C
AC           C

# dummy atom types
MW           DUMMY

# other types are guessed from the name; see
# topology.core.guess_atom_elements()
"""

#: Dictionary with hard-coded special atom names, used for guessing atom types
#: with :func:`MDAnalysis.topology.core.guess_atom_type`.
atomelements = kv2dict(TABLE_ATOMELEMENTS)

# fmt: off
elements = ['H',
            'LI', 'BE', 'B', 'C', 'N', 'O', 'F',
            'NA', 'MG', 'AL', 'P', 'SI', 'S', 'CL',
            'K']
# fmt: on

#: Plain-text table with atomic masses in u.
TABLE_MASSES = """
# masses for elements in atomic units (u)
# (taken from CHARMM and Gromacs atommass.dat)

#------------ -----------
# atomtype    mass
#------------ -----------
Ac    227.028
Al    26.981539
Am    243
Sb    121.757
Ar    39.948
As    74.92159
At    210
Ba    137.327
Bk    247
Be    9.012182
Bi    208.98037
Bh    264
B     10.811
BR    79.90400
Cd    112.411
CA    40.08000
Cf    251
C     12.01100
Ce    140.11600
CS    132.90000
CL    35.45000
Cr    51.9961
Co    58.9332
CU    63.54600
Cm    247
Db    262
Dy    162.5
Es    252
Er    167.26
Eu    151.965
Fm    257
F     18.99800
Fr    223
Gd    157.25
Ga    69.723
Ge    72.61
Au    196.96654
Hf    178.49
Hs    265
HE    4.00260
Ho    164.93032
H     1.00800
In    114.82
I     126.90450
Ir    192.22
FE    55.84700
Kr    83.8
La    138.9055
Lr    262
Pb    207.2
Li    6.941
Lu    174.967
MG    24.30500
Mn    54.93805
Mt    266
Md    258
Hg    200.59
Mo    95.94
N     14.00700
NA    22.98977
Nd    144.24
NE    20.17970
Np    237.048
Ni    58.6934
Nb    92.90638
No    259
Os    190.2
O     15.99900
Pd    106.42
P     30.97400
Pt    195.08
Pu    244
Po    209
K     39.10200
Pr    140.90765
Pm    145
Pa    231.0359
Ra    226.025
Rn    222
Re    186.207
Rh    102.9055
RB    85.46780
Ru    101.07
Rf    261
Sm    150.36
Sc    44.95591
Sg    263
Se    78.96
Si    28.0855
Ag    107.8682
Na    22.989768
Sr    87.62
S     32.06000
Ta    180.9479
Tc    98
Te    127.6
Tb    158.92534
Tl    204.3833
Th    232.0381
Tm    168.93421
Sn    118.71
Ti    47.88
W     183.85
U     238.0289
V     50.9415
Xe    131.29
Yb    173.04
Y     88.90585
ZN    65.37000
Zr    91.224
DUMMY 0.0
"""

#: Dictionary table with atomic masses in u, indexed by the element from
#: :data:`atomelements`.
masses = kv2dict(TABLE_MASSES, convertor=float)

#: Plain-text table with vdw radii.
TABLE_VDWRADII = r"""
# Van der Waals radii taken from
# [1] Bondi, A. (1964). "Van der Waals Volumes and Radii".
#     J. Phys. Chem. 68 (3): 441-451. doi:10.1021/j100785a001.
# [2] Rowland and Taylor (1996). "Intermolecular Nonbonded Contact Distances in Organic Crystal Structures:
#                                 Comparison with Distances Expected from van der Waals Radii".
#     J. Phys. Chem., 1996, 100 (18), 7384.7391. doi:10.1021/jp953141+.
# [3] Mantina, et al. (2009). "Consistent van der Waals Radii for the Whole Main Group".
#     J. Phys. Chem. A, 2009, 113 (19), 5806-5812. doi:10.1021/jp8111556.
#------------ -----------
# atomtype    r_vdw
#------------ -----------
H    1.10
HE   1.40
LI   1.82
BE   1.53
B    1.92
C    1.70
N    1.55
O    1.52
F    1.47
NE   1.54
NA   2.27
MG   1.73
AL   1.84
SI   2.10
P    1.80
S    1.80
CL   1.75
AR   1.88
K    2.75
CA   2.31
NI   1.63
CU   1.40
ZN   1.39
GA   1.87
GE   2.11
AA   1.85
SE   1.90
BR   1.85
KR   2.02
RR   3.03
SR   2.49
PD   1.63
AG   1.72
CD   1.58
IN   1.93
SN   2.17
SB   2.06
TE   2.06
I    1.98
XE   2.16
CS   3.43
BA   2.68
PT   1.75
AU   1.66
HH   1.55
TL   1.96
PB   2.02
BI   2.07
PO   1.97
AT   2.02
RN   2.20
FR   3.48
RA   2.83
U    1.86
"""

#: Dictionary table with vdw radii, indexed by the element from
#: :data:`atomelements`.
#: .. SeeAlso:: :func:`MDAnalysis.topology.core.guess_bonds`
vdwradii = kv2dict(TABLE_VDWRADII, convertor=float)

# fmt: off
Z2SYMB = {1: 'H',                                                                2: 'He',
          3: 'Li',   4: 'Be',  5: 'B',   6: 'C',   7: 'N',   8: 'O',   9: 'F',  10: 'Ne',
          11: 'Na', 12: 'Mg', 13: 'Al', 14: 'Si', 15: 'P',  16: 'S',  17: 'Cl', 18: 'Ar',
          19: 'K',  20: 'Ca', 21: 'Sc', 22: 'Ti', 23: 'V',  24: 'Cr', 25: 'Mn', 26: 'Fe',
          27: 'Co', 28: 'Ni', 29: 'Cu', 30: 'Zn', 31: 'Ga', 32: 'Ge', 33: 'As', 34: 'Se',
          35: 'Br', 36: 'Kr', 37: 'Rb', 38: 'Sr', 39: 'Y',  40: 'Zr', 41: 'Nb', 42: 'Mo',
          43: 'Tc', 44: 'Ru', 45: 'Rh', 46: 'Pd', 47: 'Ag', 48: 'Cd', 49: 'In', 50: 'Sn',
          51: 'Sb', 52: 'Te', 53: 'I',  54: 'Xe', 55: 'Cs', 56: 'Ba', 57: 'La', 58: 'Ce',
          59: 'Pr', 60: 'Nd', 61: 'Pm', 62: 'Sm', 63: 'Eu', 64: 'Gd', 65: 'Tb', 66: 'Dy',
          67: 'Ho', 68: 'Er', 69: 'Tm', 70: 'Yb', 71: 'Lu', 72: 'Hf', 73: 'Ta', 74: 'W',
          75: 'Re', 76: 'Os', 77: 'Ir', 78: 'Pt', 79: 'Au', 80: 'Hg', 81: 'Tl', 82: 'Pb',
          83: 'Bi', 84: 'Po', 85: 'At', 86: 'Rn', 87: 'Fr', 88: 'Ra', 89: 'Ac', 90: 'Th',
          91: 'Pa', 92: 'U',  93: 'Np', 94: 'Pu', 95: 'Am', 96: 'Cm', 97: 'Bk', 98: 'Cf',
          99: 'Es', 100: 'Fm', 101: 'Md', 102: 'No', 103: 'Lr', 104: 'Rf', 105: 'Db',
          106: 'Sg', 107: 'Bh', 108: 'Hs', 109: 'Mt', 110: 'Ds', 111: 'Rg', 112: 'Cn',
          113: 'Nh', 114: 'Fl', 115: 'Mc', 116: 'Lv', 117: 'Ts', 118: 'Og'}
# fmt: on

SYMB2Z = {v: k for k, v in Z2SYMB.items()}

# Conversion between SYBYL atom types and corresponding elements
# Tripos MOL2 file format:
#   https://web.archive.org/web/*/http://chemyang.ccnu.edu.cn/ccb/server/AIMMS/mol2.pdf
# fmt: off
SYBYL2SYMB = {
    "H": "H", "H.spc": "H", "H.t3p": "H",
    "C.3": "C", "C.2": "C", "C.1": "C", "C.ar": "C", "C.cat": "C",
    "N.3": "N", "N.2": "N", "N.1": "N", "N.ar": "N",
    "N.am": "N", "N.pl3": "N", "N.4": "N",
    "O.3": "O", "O.2": "O", "O.co2": "O", "O.spc": "O", "O.t3p": "O",
    "S.3": "S", "S.2": "S", "S.O": "S", "S.O2": "S",
    "S.o": "S", "S.o2": "S",  # Non-standard but often found in the wild...
    "P.3": "P",
    "F": "F",
    "Li": "Li",
    "Na": "Na",
    "Mg": "Mg",
    "Al": "Al",
    "Si": "Si",
    "K": "K",
    "Ca": "Ca",
    "Cr.th": "Cr",
    "Cr.oh": "Cr",
    "Mn": "Mn",
    "Fe": "Fe",
    "Co.oh": "Co",
    "Cu": "Cu",
    "Cl": "Cl",
    "Br": "Br",
    "I": "I",
    "Zn": "Zn",
    "Se": "Se",
    "Mo": "Mo",
    "Sn": "Sn",
}
# fmt: on



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""NAMDBIN files format --- :mod:`MDAnalysis.coordinates.NAMDBIN`
================================================================================

Read/Write coordinates in `NAMD double-precision binary format`_ (suffix "coor" or "namdbin").

.. _`NAMD double-precision binary format` : https://www.ks.uiuc.edu/Research/namd/2.10/ug/node11.html#SECTION00061500000000000000


Classes
-------

.. autoclass:: NAMDBINReader
   :members:

.. autoclass:: NAMDBINWriter
   :members:

"""
from struct import pack
import numpy as np

from . import base
from ..lib import util


class NAMDBINReader(base.SingleFrameReaderBase):
    """Reader for NAMD binary coordinate files.


    .. versionadded:: 1.0.0
    """

    format = ["COOR", "NAMDBIN"]
    units = {"length": "Angstrom"}

    def _read_first_frame(self):
        # Read header
        with open(self.filename, "rb") as namdbin:
            self.n_atoms = np.fromfile(namdbin, dtype=np.int32, count=1)[0]
            self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)
            self.ts.frame = 0
            coord_double = np.fromfile(
                namdbin, dtype=np.float64, count=self.n_atoms * 3
            )
            self.ts._pos[:] = np.array(coord_double, float).reshape(
                self.n_atoms, 3
            )

    @staticmethod
    def parse_n_atoms(filename, **kwargs):
        with open(filename, "rb") as namdbin:
            n_atoms = np.fromfile(namdbin, dtype=np.int32, count=1)[0]
        return n_atoms

    def Writer(self, filename, **kwargs):
        """Returns a NAMDBINWriter for *filename*.

        Parameters
        ----------
        filename: str
            filename of the output NAMDBIN file

        Returns
        -------
        :class:`NAMDBINWriter`

        """
        return NAMDBINWriter(filename, **kwargs)


class NAMDBINWriter(base.WriterBase):
    """Writer for NAMD binary coordinate files.


    Note
    ----
    * Does not handle writing to bz2 or gz compressed file types.


    .. versionadded:: 1.0.0
    """

    format = ["COOR", "NAMDBIN"]
    units = {"time": None, "length": "Angstrom"}

    def __init__(self, filename, n_atoms=None, **kwargs):
        """
        Parameters
        ----------
        filename : str or :class:`~MDAnalysis.lib.util.NamedStream`
             name of the output file or a stream
        n_atoms  : int
            number of atoms for the output coordinate
        """
        self.filename = util.filename(filename)

    def _write_next_frame(self, obj):
        """Write information associated with ``obj`` at current frame into
        trajectory


        Parameters
        ----------
        obj : :class:`~MDAnalysis.core.groups.AtomGroup` or
              :class:`~MDAnalysis.core.universe.Universe`
              write coordinate information associated with `obj`


        .. versionchanged:: 1.0.0
           Renamed from `write` to `_write_next_frame`.
        .. versionchanged:: 2.0.0
           Deprecated support for Timestep argument has now been removed.
           Use AtomGroup or Universe as an input instead.
        """
        if hasattr(obj, "atoms"):  # AtomGroup or Universe
            atoms = obj.atoms
            n_atoms = len(atoms)
            coor = atoms.positions.reshape(n_atoms * 3)
        else:
            errmsg = "Input obj is neither an AtomGroup or Universe"
            raise TypeError(errmsg) from None

        with util.openany(self.filename, "wb") as namdbin:
            # Write NUMATOMS
            namdbin.write(pack("i", n_atoms))
            # Write Coordinate
            namdbin.write(pack("{:d}d".format(len(coor)), *coor))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""GSD topology parser
=========================

.. versionadded:: 0.17.0

The :class:`GSDParser` generates a topology from HOOMD_ GSD topology/trajectory
files. The GSD file stores information on both the topology and the trajectory
in the same file, and allows for varying atom numbers/identities and topologies
during the course of the simulation. At the moment MDAnalysis can deal only
with the case in which there is no variation. The trajectory data are read with
the :class:`~MDAnalysis.coordinates.GSD.GSDReader` class.

.. _HOOMD: http://codeblue.umich.edu/hoomd-blue/index.html
.. _HOOMD GSD: https://github.com/glotzerlab/gsd


To load a GSD HOOMD file::

   import MDAnalysis as mda
   u = mda.Universe("example.gsd")


Classes
-------

.. autoclass:: GSDParser
   :members:
   :inherited-members:

"""
import os
import numpy as np

from .base import TopologyReaderBase
from ..core.topology import Topology
from ..core.topologyattrs import (
    Atomtypes,
    Atomnames,
    Atomids,
    Angles,
    Bonds,
    Charges,
    Dihedrals,
    Impropers,
    Masses,
    Radii,
    Resids,
    Resnums,
    Resnames,
    Segids,
)

try:
    import gsd.hoomd
except ImportError:
    HAS_GSD = False
else:
    HAS_GSD = True


class GSDParser(TopologyReaderBase):
    """Parses a Hoomd GSD file to create a Topology

    Reads the following Attributes:
     - Atomtypes
     - Bonds
     - Angles
     - Dihedrals
     - Impropers
     - Radii
     - Masses

    The GSD file also stores a `body` property in the particles, and the parser
    uses this information to set the residue names and indices.

    NOTE: if the `body` index of any particle is negative, the parser will add
    an integer number (the absolute value of the minimum of all the body
    indices) to all the body indices. This is because MDAnalysis cannot handle
    negative residue indices. This means that in that case the residue index in
    the MDAnalysis.Universe will not correspond to the body index stored in the
    GSD file.

    """

    format = "GSD"

    def __init__(self, filename):

        if not HAS_GSD:
            errmsg = (
                "GSDParser: To read a Topology from a Hoomd GSD "
                "file, please install gsd"
            )
            raise ImportError(errmsg)
        super(GSDParser, self).__init__(filename)

    def parse(self, **kwargs):
        """Parse Hoomd GSD file

        .. versionadded:: 0.17.0
        """
        attrs = {}

        with gsd.hoomd.open(self.filename, mode="r") as t:
            # Here it is assumed that the particle data does not change in the
            # trajectory.
            snap = t[0]

            natoms = snap.particles.N

            ptypes = snap.particles.types
            atypes = [ptypes[idx] for idx in snap.particles.typeid]
            if len(atypes) != natoms:
                raise IOError("Number of types does not equal natoms.")
            attrs["types"] = Atomtypes(np.array(atypes, dtype=object))

            # set radii, masses, charges
            p = snap.particles
            attrs["diameter"] = Radii(
                np.array(p.diameter / 2.0, dtype=np.float32)
            )
            attrs["mass"] = Masses(np.array(p.mass, dtype=np.float64))
            attrs["charge"] = Charges(np.array(p.charge, dtype=np.float32))

            # set bonds, angles, dihedrals, impropers
            for (
                attrname,
                attr,
            ) in (
                ("bonds", Bonds),
                ("angles", Angles),
                ("dihedrals", Dihedrals),
                ("impropers", Impropers),
            ):
                try:
                    val = getattr(snap, attrname)
                    vals = [tuple(b_instance) for b_instance in val.group]
                except:
                    vals = []
                attrs[attrname] = attr(vals)

            # get body ids to set residue number and ids
            blist = snap.particles.body.astype(np.int64)
            bodies = np.unique(blist).astype(np.int32)
            # this fixes the fact that the Topology constructor gets stuck in an
            # infinite loop if any resid is negative.
            if (blist < 0).any():
                m = blist.min()
                blist += abs(m)
            bodies = np.unique(blist).astype(np.int32)
            nbodies = bodies.size

        attrs = list(attrs.values())
        attrs.append(Atomnames(np.array(atypes, dtype=object)))
        attrs.append(Atomids(np.arange(natoms) + 1))
        attrs.append(Resids(bodies))
        attrs.append(Resnums(bodies))
        attrs.append(Resnames(bodies))
        attrs.append(Segids(np.array(["SYSTEM"], dtype=object)))

        top = Topology(natoms, nbodies, 1, attrs=attrs, atom_resindex=blist)

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Linear Density --- :mod:`MDAnalysis.analysis.lineardensity`
===========================================================

A tool to compute mass and charge density profiles along the three
cartesian axes [xyz] of the simulation cell. Works only for orthorombic,
fixed volume cells (thus for simulations in canonical NVT ensemble).
"""
import os.path as path

import numpy as np
import warnings

from MDAnalysis.analysis.base import AnalysisBase, Results
from MDAnalysis.units import constants
from MDAnalysis.lib.util import deprecate
from MDAnalysis.analysis.results import ResultsGroup


# TODO: Remove in version 3.0.0
class Results(Results):
    """From version 3.0.0 onwards, some entries in Results will be renamed. See
    the docstring for LinearDensity for details. The Results class is defined
    here to implement deprecation warnings for the user."""

    _deprecation_dict = {
        "pos": "mass_density",
        "pos_std": "mass_density_stddev",
        "char": "charge_density",
        "char_std": "charge_density_stddev",
    }

    def _deprecation_warning(self, key):
        warnings.warn(
            f"`{key}` is deprecated and will be removed in version 3.0.0. "
            f"Please use `{self._deprecation_dict[key]}` instead.",
            DeprecationWarning,
        )

    def __getitem__(self, key):
        if key in self._deprecation_dict.keys():
            self._deprecation_warning(key)
            return super(Results, self).__getitem__(
                self._deprecation_dict[key]
            )
        return super(Results, self).__getitem__(key)

    def __getattr__(self, attr):
        if attr in self._deprecation_dict.keys():
            self._deprecation_warning(attr)
            attr = self._deprecation_dict[attr]
        return super(Results, self).__getattr__(attr)


class LinearDensity(AnalysisBase):
    r"""Linear density profile

    Parameters
    ----------
    select : AtomGroup
          any atomgroup
    grouping : str {'atoms', 'residues', 'segments', 'fragments'}
          Density profiles will be computed either on the atom positions (in
          the case of 'atoms') or on the center of mass of the specified
          grouping unit ('residues', 'segments', or 'fragments').
    binsize : float
          Bin width in Angstrom used to build linear density
          histograms. Defines the resolution of the resulting density
          profile (smaller --> higher resolution)
    verbose : bool, optional
          Show detailed progress of the calculation if set to ``True``

    Attributes
    ----------
    results.x.dim : int
           index of the [xyz] axes
    results.x.mass_density : numpy.ndarray
           mass density in :math:`g \cdot cm^{-3}` in [xyz] direction
    results.x.mass_density_stddev : numpy.ndarray
           standard deviation of the mass density in [xyz] direction
    results.x.charge_density : numpy.ndarray
           charge density in :math:`\mathrm{e} \cdot mol \cdot cm^{-3}` in
           [xyz] direction
    results.x.charge_density_stddev : numpy.ndarray
           standard deviation of the charge density in [xyz] direction
    results.x.pos: numpy.ndarray
        Alias to the :attr:`results.x.mass_density` attribute.

        .. deprecated:: 2.2.0
           Will be removed in MDAnalysis 3.0.0. Please use
           :attr:`results.x.mass_density` instead.
    results.x.pos_std: numpy.ndarray
        Alias to the :attr:`results.x.mass_density_stddev` attribute.

        .. deprecated:: 2.2.0
           Will be removed in MDAnalysis 3.0.0. Please use
           :attr:`results.x.mass_density_stddev` instead.
    results.x.char: numpy.ndarray
        Alias to the :attr:`results.x.charge_density` attribute.

        .. deprecated:: 2.2.0
           Will be removed in MDAnalysis 3.0.0. Please use
           :attr:`results.x.charge_density` instead.
    results.x.char_std: numpy.ndarray
        Alias to the :attr:`results.x.charge_density_stddev` attribute.

        .. deprecated:: 2.2.0
           Will be removed in MDAnalysis 3.0.0. Please use
           :attr:`results.x.charge_density_stddev` instead.
    results.x.slice_volume : float
           volume of bin in [xyz] direction
    results.x.hist_bin_edges : numpy.ndarray
           edges of histogram bins for mass/charge densities, useful for, e.g.,
           plotting of histogram data.
    Note: These density units are likely to be changed in the future.

    Example
    -------
    First create a :class:`LinearDensity` object by supplying a selection,
    then use the :meth:`run` method. Finally access the results
    stored in results, i.e. the mass density in the x direction.

    .. code-block:: python

       ldens = LinearDensity(selection)
       ldens.run()
       print(ldens.results.x.mass_density)


    Alternatively, other types of grouping can be selected using the
    ``grouping`` keyword. For example to calculate the density based on
    a grouping of the :class:`~MDAnalysis.core.groups.ResidueGroup`
    of the input :class:`~MDAnalysis.core.groups.AtomGroup`.

    .. code-block:: python

       ldens = LinearDensity(selection, grouping='residues', binsize=1.0)
       ldens.run()



    .. versionadded:: 0.14.0

    .. versionchanged:: 1.0.0
       Support for the ``start``, ``stop``, and ``step`` keywords has been
       removed. These should instead be passed to :meth:`LinearDensity.run`.
       The ``save()`` method was also removed, you can use ``np.savetxt()`` or
       ``np.save()`` on the :attr:`LinearDensity.results` dictionary contents
       instead.

    .. versionchanged:: 1.0.0
       Changed `selection` keyword to `select`

    .. versionchanged:: 2.0.0
       Results are now instances of
       :class:`~MDAnalysis.core.analysis.Results` allowing access
       via key and attribute.

    .. versionchanged:: 2.2.0

       *  Fixed a bug that caused LinearDensity to fail if grouping="residues"
          or grouping="segments" were set.
       *  Residues, segments, and fragments will be analysed based on their
          centre of mass, not centre of geometry as previously stated.
       *  LinearDensity now works with updating atom groups.
       *  Added new result container :attr:`results.x.hist_bin_edges`.
          It contains the bin edges of the histrogram bins for calculated
          densities and can be used for easier plotting of histogram data.

    .. versionchanged:: 2.10.0
       *  Introduced :meth:`get_supported_backends` allowing for parallel execution
          on :mod:`multiprocessing` and :mod:`dask` backends.
       *  Removed undocumented and unused attribute :attr:`totalmass`.

    .. deprecated:: 2.2.0
       The `results` dictionary has been changed and the attributes
       :attr:`results.x.pos`, :attr:`results.x.pos_std`, :attr:`results.x.char`
       and :attr:`results.x.char_std` are now deprecated. They will be removed
       in 3.0.0. Please use :attr:`results.x.mass_density`,
       :attr:`results.x.mass_density_stddev`, :attr:`results.x.charge_density`,
       and :attr:`results.x.charge_density_stddev` instead.
    """

    _analysis_algorithm_is_parallelizable = True

    @classmethod
    def get_supported_backends(cls):
        return (
            "serial",
            "multiprocessing",
            "dask",
        )

    def __init__(self, select, grouping="atoms", binsize=0.25, **kwargs):
        super(LinearDensity, self).__init__(
            select.universe.trajectory, **kwargs
        )
        # allows use of run(parallel=True)
        self._ags = [select]
        self._universe = select.universe

        self.binsize = binsize

        # group of atoms on which to compute the COM (same as used in
        # AtomGroup.wrap())
        self.grouping = grouping

        # Initiate result instances
        self.results["x"] = Results(dim=0)
        self.results["y"] = Results(dim=1)
        self.results["z"] = Results(dim=2)

        # Box sides
        self.dimensions = self._universe.dimensions[:3]
        self.volume = np.prod(self.dimensions)
        # number of bins
        bins = (self.dimensions // self.binsize).astype(int)

        # Here we choose a number of bins of the largest cell side so that
        # x, y and z values can use the same "coord" column in the output file
        self.nbins = bins.max()
        slices_vol = self.volume / bins

        self.keys = [
            "mass_density",
            "mass_density_stddev",
            "charge_density",
            "charge_density_stddev",
        ]

        # Initialize results array with zeros
        for dim in self.results:
            idx = self.results[dim]["dim"]
            self.results[dim]["slice_volume"] = slices_vol[idx]
            for key in self.keys:
                self.results[dim][key] = np.zeros(self.nbins)

        # Get masses and charges for the selection (e.g. UpdatingAtomGroup)
        if self.grouping == "atoms":
            self.masses = self._ags[0].masses
            self.charges = self._ags[0].charges

        elif self.grouping in ["residues", "segments", "fragments"]:
            self.masses = self._ags[0].total_mass(compound=self.grouping)
            self.charges = self._ags[0].total_charge(compound=self.grouping)

        else:
            raise AttributeError(
                f"{self.grouping} is not a valid value for grouping."
            )

    @staticmethod
    def _custom_aggregator(results):
        # NB: the *stddev values here are not the standard deviation,
        # but the variance. The stddev is calculated in _conclude()
        mass_density = np.sum(
            [entry["mass_density"] for entry in results], axis=0
        )
        mass_density_stddev = np.sum(
            [entry["mass_density_stddev"] for entry in results], axis=0
        )
        charge_density = np.sum(
            [entry["charge_density"] for entry in results], axis=0
        )
        charge_density_stddev = np.sum(
            [entry["charge_density_stddev"] for entry in results], axis=0
        )
        return Results(
            dim=results[0]["dim"],
            slice_volume=results[0]["slice_volume"],
            hist_bin_edges=results[0]["hist_bin_edges"],
            mass_density=mass_density,
            mass_density_stddev=mass_density_stddev,
            charge_density=charge_density,
            charge_density_stddev=charge_density_stddev,
        )

    def _get_aggregator(self):
        return ResultsGroup(
            lookup={
                "x": self._custom_aggregator,
                "y": self._custom_aggregator,
                "z": self._custom_aggregator,
            }
        )

    def _single_frame(self):
        if self.grouping == "atoms":
            self.masses = self._ags[0].masses
            self.charges = self._ags[0].charges

        elif self.grouping in ["residues", "segments", "fragments"]:
            self.masses = self._ags[0].total_mass(compound=self.grouping)
            self.charges = self._ags[0].total_charge(compound=self.grouping)

        else:
            raise AttributeError(
                f"{self.grouping} is not a valid value for grouping."
            )

        self.group = getattr(self._ags[0], self.grouping)
        self._ags[0].wrap(compound=self.grouping)
        # Find position of atom/group of atoms
        if self.grouping == "atoms":
            positions = self._ags[0].positions  # faster for atoms
        else:
            # Centre of mass for residues, segments, fragments
            positions = self._ags[0].center_of_mass(compound=self.grouping)

        for dim in ["x", "y", "z"]:
            idx = self.results[dim]["dim"]

            key = "mass_density"
            key_std = "mass_density_stddev"
            # histogram for positions weighted on masses
            hist, _ = np.histogram(
                positions[:, idx],
                weights=self.masses,
                bins=self.nbins,
                range=(0.0, max(self.dimensions)),
            )

            self.results[dim][key] += hist
            self.results[dim][key_std] += np.square(hist)

            key = "charge_density"
            key_std = "charge_density_stddev"
            # histogram for positions weighted on charges
            hist, bin_edges = np.histogram(
                positions[:, idx],
                weights=self.charges,
                bins=self.nbins,
                range=(0.0, max(self.dimensions)),
            )

            self.results[dim][key] += hist
            self.results[dim][key_std] += np.square(hist)
            self.results[dim]["hist_bin_edges"] = bin_edges

    def _conclude(self):
        avogadro = constants["N_Avogadro"]  # unit: mol^{-1}
        volume_conversion = 1e-24  # unit: A^3/cm^3
        # divide result values by avodagro and convert from A3 to cm3
        k = avogadro * volume_conversion

        # Average results over the number of configurations
        for dim in ["x", "y", "z"]:
            for key in [
                "mass_density",
                "mass_density_stddev",
                "charge_density",
                "charge_density_stddev",
            ]:
                self.results[dim][key] /= self.n_frames
            # Compute standard deviation for the error
            # For certain tests in testsuite, floating point imprecision
            # can lead to negative radicands of tiny magnitude (yielding nan).
            # radicand_mass and radicand_charge are therefore calculated first
            # and negative values set to 0 before the square root
            # is calculated.
            radicand_mass = self.results[dim][
                "mass_density_stddev"
            ] - np.square(self.results[dim]["mass_density"])
            radicand_mass[radicand_mass < 0] = 0
            self.results[dim]["mass_density_stddev"] = np.sqrt(radicand_mass)

            radicand_charge = self.results[dim][
                "charge_density_stddev"
            ] - np.square(self.results[dim]["charge_density"])
            radicand_charge[radicand_charge < 0] = 0
            self.results[dim]["charge_density_stddev"] = np.sqrt(
                radicand_charge
            )

        for dim in ["x", "y", "z"]:
            # norming factor, units of mol^-1 cm^3
            norm = k * self.results[dim]["slice_volume"]
            for key in self.keys:
                self.results[dim][key] /= norm

    # TODO: Remove in 3.0.0
    @deprecate(
        release="2.2.0",
        remove="3.0.0",
        message="It will be replaced by a :meth:`_reduce` "
        "method in the future",
    )
    def _add_other_results(self, other):
        """For parallel analysis"""
        for dim in ["x", "y", "z"]:
            for key in self.keys:
                self.results[dim][key] += other.results[dim][key]
